import argparse
import asyncio
import datetime
import importlib
import sys
from typing import Any, Dict, List, Optional

import structlog

import rasa.cli.utils
import rasa.shared.utils.cli
from rasa.cli import SubParsersAction
from rasa.cli.arguments.default_arguments import (
    add_endpoint_param,
    add_model_param,
    add_remote_storage_param,
)
from rasa.core.agent import Agent
from rasa.core.exceptions import AgentNotReady
from rasa.core.processor import MessageProcessor
from rasa.core.utils import AvailableEndpoints
from rasa.dialogue_understanding.commands import Command
from rasa.dialogue_understanding.generator import (
    LLMBasedCommandGenerator,
)
from rasa.dialogue_understanding.generator.command_parser import DEFAULT_COMMANDS
from rasa.dialogue_understanding_test.command_metric_calculation import (
    calculate_command_metrics,
)
from rasa.dialogue_understanding_test.constants import (
    DEFAULT_INPUT_TESTS_PATH,
    KEY_STUB_CUSTOM_ACTIONS,
)
from rasa.dialogue_understanding_test.du_test_result import (
    DialogueUnderstandingTestResult,
    DialogueUnderstandingTestSuiteResult,
)
from rasa.dialogue_understanding_test.du_test_runner import (
    DialogueUnderstandingTestRunner,
)
from rasa.dialogue_understanding_test.io import (
    read_test_suite,
    write_test_results_to_file,
)
from rasa.dialogue_understanding_test.validation import (
    validate_cli_arguments,
    validate_test_cases,
)
from rasa.e2e_test.e2e_test_case import TestSuite
from rasa.exceptions import RasaException
from rasa.shared.constants import (
    DEFAULT_ENDPOINTS_PATH,
    LLM_CONFIG_KEY,
    ROUTE_TO_CALM_SLOT,
)
from rasa.shared.core.domain import Domain
from rasa.shared.core.flows import FlowsList
from rasa.shared.utils.llm import (
    combine_custom_and_default_config,
    resolve_model_client_config,
)
from rasa.utils.beta import ensure_beta_feature_is_enabled
from rasa.utils.endpoints import EndpointConfig

RASA_PRO_BETA_DIALOGUE_UNDERSTANDING_TEST_ENV_VAR_NAME = (
    "RASA_PRO_BETA_DIALOGUE_UNDERSTANDING_TEST"
)

structlogger = structlog.get_logger()


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add the dialogue understanding test subparser to `rasa test`.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    for subparser in subparsers.choices.values():
        if subparser.prog == "rasa test":
            du_test_subparser = create_du_test_subparser(parents)

            for action in subparser._subparsers._actions:
                if action.choices is not None:
                    action.choices["du"] = du_test_subparser
                    return

    # If we get here, we couldn't hook the subparser to `rasa test`
    raise RasaException(
        "Hooking the dialogue understanding (du) test subparser to "
        "`rasa test` command could not be completed. "
        "Cannot run dialogue understanding testing."
    )


def create_du_test_subparser(
    parents: List[argparse.ArgumentParser],
) -> argparse.ArgumentParser:
    """Create dialogue understanding test subparser."""
    du_test_subparser = argparse.ArgumentParser(
        prog="rasa test du",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Runs dialogue understanding testing.",
    )

    du_test_subparser.set_defaults(func=execute_dialogue_understanding_tests)

    add_du_test_arguments(du_test_subparser)
    add_bot_arguments(du_test_subparser)

    return du_test_subparser


def add_bot_arguments(parser: argparse.ArgumentParser) -> None:
    bot_arguments = parser.add_argument_group("Bot Settings")
    add_model_param(bot_arguments, add_positional_arg=False)
    add_endpoint_param(
        bot_arguments,
        help_text="Configuration file for the model server and the connectors as a "
        "yml file.",
    )
    add_remote_storage_param(bot_arguments)


def add_du_test_arguments(parser: argparse.ArgumentParser) -> None:
    """Arguments for running dialogue understanding tests."""
    du_arguments = parser.add_argument_group("Testing Settings")
    du_arguments.add_argument(
        "path-to-test-cases",
        nargs="?",
        type=str,
        default=DEFAULT_INPUT_TESTS_PATH,
        help="Input file or folder containing dialogue understanding test cases.",
    )
    du_arguments.add_argument(
        "--output-file",
        type=str,
        default="dialogue_understanding_test_{date:%Y%m%d-%H%M%S}.yml".format(
            date=datetime.datetime.now()
        ),
        help="Path to the output file to write the results to.",
    )
    du_arguments.add_argument(
        "--no-output",
        action="store_true",
        help="If set, no output file will be written to disk.",
    )
    du_arguments.add_argument(
        "--output-prompt",
        action="store_true",
        help="If set, the dialogue understanding test output will contain "
        "prompts for each failure.",
    )
    du_arguments.add_argument(
        "--additional-commands",
        type=str,
        nargs="*",
        help=(
            "List of additional custom command classes to add, separated by spaces. "
            "For example: --additional-commands my_module.MyCustomCommand"
        ),
    )
    du_arguments.add_argument(
        "--remove-default-commands",
        type=str,
        nargs="*",
        help=(
            f"List of default commands to remove, separated by spaces. "
            f"Default commands include: "
            f"{', '.join([command.__name__ for command in DEFAULT_COMMANDS])}. "
            f"For example: --remove-default-commands ClarifyCommand HumanHandoffCommand"
        ),
    )


def ensure_calm_only_bot(agent: Agent) -> None:
    if agent.domain is None or agent.processor is None:
        return

    if ROUTE_TO_CALM_SLOT in [slot.name for slot in agent.domain.slots]:
        rasa.shared.utils.cli.print_error(
            "You are using coexistence. Dialogue Understanding Tests do only work for "
            "CALM only assistants."
        )
        sys.exit(0)

    if not agent.processor.is_calm_assistant:
        rasa.shared.utils.cli.print_error(
            "Dialogue Understanding Tests do only work for CALM assistants. "
            "Your assistant does not use CALM."
        )
        sys.exit(0)


def execute_dialogue_understanding_tests(args: argparse.Namespace) -> None:
    """Run the dialogue understanding tests.

    Args:
        args: Commandline arguments.
    """
    ensure_beta_feature_is_enabled(
        "Dialogue Understanding (DU) Testing",
        env_flag=RASA_PRO_BETA_DIALOGUE_UNDERSTANDING_TEST_ENV_VAR_NAME,
    )

    # basic validation of the passed CLI arguments
    validate_cli_arguments(args)

    # initialization of endpoints
    endpoints = set_up_available_endpoints(args)

    # set up the test runner, e.g. start the agent
    try:
        test_runner = DialogueUnderstandingTestRunner(
            endpoints=endpoints,
            model_path=args.model,
            model_server=endpoints.model,
            remote_storage=args.remote_storage,
        )
    except AgentNotReady as error:
        structlogger.error(
            "rasa.dialogue_understanding_test.agent_not_ready", message=error.message
        )
        sys.exit(1)

    # Exit if the bot is not calm only
    ensure_calm_only_bot(test_runner.agent)

    # Ensure processor is not None so that we can extract the flows and the llm config
    if test_runner.agent.processor is None:
        rasa.shared.utils.cli.print_error(
            "No processor: Not able to retrieve flows and config from trained model."
        )
        sys.exit(0)

    # flows are needed in order to parse the commands when reading the test cases
    flows = asyncio.run(test_runner.agent.processor.get_flows())
    # llm config is needed for instrumentation
    llm_config = _get_llm_command_generator_config(test_runner.agent.processor)

    # read test cases from the given path
    test_suite: TestSuite = get_valid_test_suite(args, flows, test_runner.agent.domain)

    # setup stub custom actions if they are used
    set_up_stub_custom_actions(test_suite, endpoints)

    # run the actual test cases
    test_results = asyncio.run(
        test_runner.run_test_cases(
            test_suite.test_cases, test_suite.fixtures, test_suite.metadata
        )
    )

    # evaluate test results
    passing_test_results, failing_test_results = split_test_results(test_results)
    command_metrics = calculate_command_metrics(test_results)

    test_suite_result = DialogueUnderstandingTestSuiteResult.from_results(
        failing_test_results, passing_test_results, command_metrics, llm_config
    )

    # Do not move this import to the top of the file as it will break the
    # instrumentation of this function: the CLI module is initialized before the
    # instrumentation is set up, and we won't be able to "replace" the function
    # with the instrumented wrapper
    from rasa.dialogue_understanding_test.io import print_test_results

    # write results to console and file
    print_test_results(test_suite_result, output_prompt=args.output_prompt)
    if not args.no_output:
        write_test_results_to_file(
            test_suite_result,
            args.output_file,
            args.output_prompt,
        )


def _import_custom_command_class(class_path: str) -> Command:
    """Dynamically import a command class from a string path."""
    try:
        module_name, class_name = class_path.rsplit(".", 1)
        module = importlib.import_module(module_name)
        clz = getattr(module, class_name)
    except (ImportError, AttributeError, ValueError) as e:
        raise ValueError(f"Failed to import class '{class_path}': {e}")
    if not issubclass(clz, Command):
        structlogger.error(
            "rasa.dialogue_understanding_test.invalid_additional_command",
            event_info="The custom command class must be a subclass of Command.",
            class_path=class_path,
        )
        sys.exit(1)
    return clz


def _extract_additional_command_classes_from_cli_args(
    args: argparse.Namespace,
) -> List[Command]:
    """Extract additional command classes from the CLI arguments."""
    additional_commands = getattr(args, "additional_commands", [])
    if not additional_commands:
        return []
    return [
        _import_custom_command_class(command_module)
        for command_module in additional_commands
    ]


def get_valid_test_suite(
    args: argparse.Namespace, flows: FlowsList, domain: Optional[Domain]
) -> TestSuite:
    """Read the test cases from the given test case path and validate them."""
    path_to_test_cases = getattr(args, "path-to-test-cases", DEFAULT_INPUT_TESTS_PATH)
    remove_default_commands = getattr(args, "remove_default_commands", [])
    custom_command_classes = _extract_additional_command_classes_from_cli_args(args)
    test_suite = read_test_suite(
        path_to_test_cases, flows, custom_command_classes, remove_default_commands
    )
    validate_test_cases(test_suite.test_cases, domain)
    return test_suite


def set_up_available_endpoints(args: argparse.Namespace) -> AvailableEndpoints:
    """Set up the available endpoints for the test runner."""
    args.endpoints = rasa.cli.utils.get_validated_path(
        args.endpoints, "endpoints", DEFAULT_ENDPOINTS_PATH, True
    )
    endpoints = AvailableEndpoints.get_instance(args.endpoints)

    # Ignore all endpoints apart from action server, model, and nlu
    # to ensure InMemoryTrackerStore is being used instead of production
    # tracker store
    endpoints.tracker_store = None
    endpoints.lock_store = None
    endpoints.event_broker = None

    # disable nlg endpoint as we don't need it for dialogue understanding tests
    endpoints.nlg = None

    return endpoints


def set_up_stub_custom_actions(
    test_suite: TestSuite, endpoints: AvailableEndpoints
) -> None:
    """Set up the stub custom actions if they are used."""
    if test_suite.stub_custom_actions:
        if not endpoints.action:
            endpoints.action = EndpointConfig()

        endpoints.action.kwargs[KEY_STUB_CUSTOM_ACTIONS] = (
            test_suite.stub_custom_actions
        )


def split_test_results(
    results: List[DialogueUnderstandingTestResult],
) -> tuple[
    List[DialogueUnderstandingTestResult], List[DialogueUnderstandingTestResult]
]:
    """Split the test results into passed and failed test cases."""
    passed_cases = [r for r in results if r.passed]
    failed_cases = [r for r in results if not r.passed]

    return passed_cases, failed_cases


def _get_llm_command_generator_config(
    processor: MessageProcessor,
) -> Optional[Dict[str, Any]]:
    from rasa.dialogue_understanding.generator.constants import DEFAULT_LLM_CONFIG

    train_schema = processor.model_metadata.train_schema

    for node_name, node in train_schema.nodes.items():
        if node.matches_type(LLMBasedCommandGenerator, include_subtypes=True):
            # Configurations can reference model groups defined in the endpoints.yml
            resolved_config = resolve_model_client_config(
                node.config.get(LLM_CONFIG_KEY, {}), node_name
            )
            return combine_custom_and_default_config(
                resolved_config, DEFAULT_LLM_CONFIG
            )

    return None



import base64
import logging
import re
import time
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Dict, Optional, Text, Union, cast

import hvac.exceptions
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.interval import IntervalTrigger

from rasa.core.secrets_manager.constants import (
    TRACKER_STORE_ENDPOINT_TYPE,
    TRANSIT_KEY_FOR_ENCRYPTION_LABEL,
    VAULT_MOUNT_POINT_DEFAULT_VALUE,
    VAULT_SECRET_MANAGER_NAME,
)
from rasa.core.secrets_manager.endpoints import (
    CredentialsLocation,
    EndpointReader,
    EndpointTrait,
    TrackerStoreEndpointValidator,
)
from rasa.core.secrets_manager.secret_manager import SecretManagerConfig, SecretsManager
from rasa.shared.exceptions import RasaException
from rasa.utils.endpoints import EndpointConfig

logger = logging.getLogger(__name__)


class VaultCredentialsLocation:
    """Represents where the secret is stored in Vault.

    Secret can be encrypted with vault's transit engine.

    source: The source of the secret.
    secret_key: The key of the secret.
    transit_key: The key used to encrypt the secret.
    """

    def __init__(
        self,
        source: Text,
        secret_key: Text,
        transit_key: Optional[Text] = None,
    ) -> None:
        """Initialise the secret.

        Args:
            source: The source of the secret.
            secret_key: The key of the secret.
            transit_key: The key used to encrypt the secret.
        """
        self.source = source
        self.secret_key = secret_key
        self.transit_key = transit_key

    @classmethod
    def from_credentials_location(
        cls, credentials_location: CredentialsLocation
    ) -> "VaultCredentialsLocation":
        """Initialise the secret from CredentialsLocation.

        Args:
            credentials_location: The CredentialsLocation.
        """
        transit_key = (
            credentials_location.kwargs.get(TRANSIT_KEY_FOR_ENCRYPTION_LABEL)
            if credentials_location.kwargs
            else None
        )

        return cls(
            source=credentials_location.source,
            secret_key=credentials_location.secret_key,
            transit_key=transit_key,
        )

    @staticmethod
    def is_vault_credential_location_instance(
        value: Union[str, "VaultCredentialsLocation"],
    ) -> bool:
        """Check if the config key is a vault encrypted secret.

        Args:
            value: The config key.

        Returns:
            True if value is an instance of VaultCredentialsLocation.
        """
        return isinstance(value, VaultCredentialsLocation)


class VaultEndpointConfigReader:
    """Used to read Vault location specification in the endpoint config."""

    def __init__(self, endpoint_config: "EndpointConfig") -> None:
        """Initialize the Vault endpoint processor.

        Args:
            endpoint_config: The endpoint config.
        """
        self._endpoint_config_reader = EndpointReader(endpoint_config=endpoint_config)

    def get_credentials_location(
        self,
        property_name: Text,
    ) -> Optional[VaultCredentialsLocation]:
        """Get the VaultCredentialsLocation for an endpoint property.

        Args:
            property_name: The name of the endpoint's property

        Returns:
            The VaultCredentialsLocation for the property.
        """
        property_value = self._endpoint_config_reader.get_property_value(property_name)

        if property_value and CredentialsLocation.is_credentials_location_instance(
            property_value
        ):
            credentials_location = cast(CredentialsLocation, property_value)
            if (
                credentials_location.get_secret_manager_name()
                == VAULT_SECRET_MANAGER_NAME
            ):
                return VaultCredentialsLocation.from_credentials_location(
                    credentials_location=credentials_location
                )
            else:
                raise RasaException(
                    f"Secret manager {credentials_location.source} is not supported."
                )

        return None

    def get_transit_keys_per_endpoint_property(self) -> Optional[Dict[Text, Text]]:
        """Get the transit keys for endpoint's properties.

        Returns:
            A dictionary of the transit keys per endpoint's properties.

        Example:
        {
            "username": "transit_key_for_username",
            "password": "transit_key_for_password"
        }

        """
        transit_keys = {}

        for config_key_name in self._endpoint_config_reader.get_keys():
            credentials_location = self.get_credentials_location(
                property_name=config_key_name
            )

            if (
                credentials_location
                and VaultCredentialsLocation.is_vault_credential_location_instance(
                    credentials_location
                )
            ):
                if credentials_location.transit_key:
                    transit_keys[credentials_location.secret_key] = (
                        credentials_location.transit_key
                    )

        return transit_keys if transit_keys else None


class VaultSecretsManager(SecretsManager):
    """Secrets Manager for Vault.

    It supports both transit secret and kv engines.
    """

    def __init__(
        self,
        host: Text,
        token: Text,
        secrets_path: Text,
        transit_mount_point: Optional[Text] = None,
        namespace: Optional[Text] = None,
        mount_point: Optional[Text] = None,
    ):
        """Initialise the VaultSecretsManager.

        Args:
            host: The host of the vault server.
            token: The token to authenticate with the vault server.
            secrets_path: The path to the secrets in the vault server.
            transit_mount_point: The mount point of the transit engine.
            namespace: The namespace in which secrets reside in.
            mount_point: The mount point of the kv engine.
        """
        self.host = host
        self.transit_mount_point = transit_mount_point
        self.token = token
        self.secrets_path = secrets_path
        self.mount_point = mount_point or VAULT_MOUNT_POINT_DEFAULT_VALUE
        self.namespace = namespace

        # Create client
        self.client = hvac.Client(
            url=self.host,
            token=self.token,
            namespace=self.namespace,
        )

        self.vault_token_manager = VaultTokenManager(host=self.host, token=self.token)

        self.vault_token_manager.start()

    def name(self) -> Text:
        """Return unique identifier of the secret manager.

        Returns:
            The unique identifier of the secret manager.
        """
        return VAULT_SECRET_MANAGER_NAME

    def load_secrets(self, endpoint_trait: "EndpointTrait") -> Dict[Text, Any]:
        """Load secrets from vault server.

        If secrets are encrypted with transit engine,
        they will be decrypted if transit_keys are provided.

        Args:
            endpoint_trait: The endpoint trait.

        Returns:
            The secrets mapped to their endpoint property.

        Example:
        {
            "username": "username stored in vault
            "password": "password stored in vault"
        }
        """
        logger.info(f"Loading secrets from vault server at {self.host}.")
        read_response = self.client.secrets.kv.read_secret_version(
            mount_point=self.mount_point, path=self.secrets_path
        )

        secrets = read_response["data"]["data"]

        reader = get_vault_endpoint_reader(endpoint_trait=endpoint_trait)

        if reader is None:
            raise RasaException(
                f"Failed to create endpoint processor "
                f"for endpoint {endpoint_trait.endpoint_type}"
            )

        transit_keys = reader.get_transit_keys_per_endpoint_property()

        if transit_keys is not None:
            self._decode_transit_secrets(secrets=secrets, transit_keys=transit_keys)

        logger.info(f"Successfully loaded secrets from vault server at {self.host}.")

        return secrets

    @staticmethod
    def _is_vault_transit_cipher(cipher: Text) -> bool:
        """Check if the cipher is a transit cipher.

        Args:
            cipher: The cipher to check.

        Returns:
            True if the cipher is a transit cipher, False otherwise.
        """
        # if cipher starts with vault:v/d: it is a transit cipher
        regexp = re.compile(r"^vault:v\d+:")
        return re.search(regexp, cipher) is not None

    def _decrypt_transit_cipher(self, cipher: Text, transit_key_name: Text) -> Text:
        """Decrypt the cipher using the transit key.

        Args:
            cipher: The cipher to decrypt.
            transit_key_name: The name of the transit key to use.

        Returns:
            The cipher decrypted with the transit key.
        """
        logger.info("Decrypting cipher using transit key.")
        decrypted_data = self.client.secrets.transit.decrypt_data(
            name=transit_key_name,
            ciphertext=cipher,
            mount_point=self.transit_mount_point,
        )
        logger.info("Finished decrypting cipher using transit key.")

        return (
            base64.standard_b64decode(decrypted_data.get("data").get("plaintext"))
            .decode("utf-8")
            .strip()
        )

    def _decode_transit_secrets(
        self, secrets: Dict[Text, Text], transit_keys: Optional[Dict[Text, Text]] = None
    ) -> None:
        """Decode values from tracker_store if they are encrypted with transit engine.

        Args:
            secrets:  Map of secrets to decode. The values will be updated in place.
            transit_keys: Map of secrets to transit keys.
        """
        if transit_keys is None:
            return

        logger.info("Start to decrypt secrets using transit secrets engine.")
        for secret_name, secret_value in secrets.items():
            if self._is_vault_transit_cipher(secret_value):
                transit_key_name = transit_keys.get(secret_name)
                if transit_key_name is not None:
                    secrets[secret_name] = self._decrypt_transit_cipher(
                        cipher=secret_value, transit_key_name=transit_key_name
                    )
        logger.info("Finished decrypting secrets using transit secrets engine.")


class VaultTokenManager:
    """VaultTokenManager renews the vault token if it renewable."""

    TOKEN_MINIMAL_REMAINING_LIFE_IN_SECONDS = 1
    TOKEN_MINIMAL_REFRESH_TIME_IN_SECONDS = 1
    # The bottom threshold for token's TTL (time to live) in seconds.
    # Set to 15 seconds to compensate for possible network latency
    TOKEN_TTL_BOTTOM_THRESHOLD_IN_SECONDS = 15
    DEFAULT_NUMBER_OF_RETRIES_FOR_TOKEN_REFRESH = 5
    TIME_TO_WAIT_BETWEEN_RETRIES_IN_SECONDS = 1

    def __init__(
        self,
        host: Text,
        token: Text,
        number_of_retries: int = DEFAULT_NUMBER_OF_RETRIES_FOR_TOKEN_REFRESH,
    ):
        """Initialise the VaultTokenManager.

        Args:
            host: The host of the vault server.
            token: The token to authenticate with the vault server.
                   If the token is expiring, it will be automatically refreshed.
            number_of_retries: The number of retries to refresh the token.
        """
        self.host = host
        self.token = token
        self.number_of_retries = number_of_retries

        self.client = hvac.Client(
            url=self.host,
            token=self.token,
        )

    def start(self) -> None:
        """Start refreshing the token if it is expiring."""
        renew_response: Dict[Text, Dict[Text, Any]] = (
            self.client.auth.token.lookup_self()
        )
        is_token_expiring = renew_response["data"]["renewable"]
        if is_token_expiring:
            refresh_interval_in_seconds = renew_response["data"]["creation_ttl"]
            first_expiring_period_in_seconds = renew_response["data"]["ttl"]
            logger.info("Token is expiring. Starting periodic token refresh.")
            self._start_token_refresh(
                first_expiring_period_in_seconds, refresh_interval_in_seconds
            )

    def _renew_token(self) -> None:
        """Renew the token."""
        index = 0
        for index in range(self.number_of_retries):
            try:
                logger.info("Renewing vault token.")
                self.client.auth.token.renew_self()
                logger.info("Finished renewing vault token.")
                break
            except Exception as e:
                if index == self.number_of_retries - 1:
                    raise RasaException(
                        f"Failed to renew vault token after "
                        f"{self.number_of_retries} retries."
                    )

                logger.warning(
                    f"Failed to renew vault token. "
                    f"Error: {e}. "
                    f"Trying again in "
                    f"{self.TIME_TO_WAIT_BETWEEN_RETRIES_IN_SECONDS} second(s)"
                )
                time.sleep(self.TIME_TO_WAIT_BETWEEN_RETRIES_IN_SECONDS)

    def _start_token_refresh(
        self, remaining_life_in_seconds: int, refresh_interval_in_seconds: int
    ) -> None:
        """Start a background job to refresh the token.

        Args:
            remaining_life_in_seconds: The remaining life of the token in seconds.
                                       This period can be shorter than the token's TTL.
            refresh_interval_in_seconds: The refresh interval in seconds.
                                         It matches the token's TTL (creation_ttl).
                                         This interval is used after the remaining life.
        """
        self.scheduler = BackgroundScheduler()
        self.scheduler.start()

        # Reduce the refresh interval and the remaining life by
        # TOKEN_TTL_BOTTOM_THRESHOLD_IN_SECONDS
        # to compensate for possible network latency.
        # If token's TTL and remaining TTL is less than
        # TOKEN_TTL_BOTTOM_THRESHOLD_IN_SECONDS
        # we must not reduce the refresh interval and the remaining life
        if refresh_interval_in_seconds > self.TOKEN_TTL_BOTTOM_THRESHOLD_IN_SECONDS:
            refresh_interval_in_seconds -= self.TOKEN_TTL_BOTTOM_THRESHOLD_IN_SECONDS
        else:
            logger.info(
                f"Token's refresh interval is less than "
                f"{self.TOKEN_TTL_BOTTOM_THRESHOLD_IN_SECONDS} seconds. "
                f"Readjusting the refresh interval to "
                f"{self.TOKEN_MINIMAL_REFRESH_TIME_IN_SECONDS} seconds."
            )
            refresh_interval_in_seconds = self.TOKEN_MINIMAL_REFRESH_TIME_IN_SECONDS

        if remaining_life_in_seconds > self.TOKEN_TTL_BOTTOM_THRESHOLD_IN_SECONDS:
            remaining_life_in_seconds -= self.TOKEN_TTL_BOTTOM_THRESHOLD_IN_SECONDS
        else:
            logger.info(
                f"Token's TTL is less than "
                f"{self.TOKEN_TTL_BOTTOM_THRESHOLD_IN_SECONDS} seconds. "
                f"Readjusting the refresh interval to "
                f"{self.TOKEN_MINIMAL_REMAINING_LIFE_IN_SECONDS} seconds."
            )
            remaining_life_in_seconds = self.TOKEN_MINIMAL_REMAINING_LIFE_IN_SECONDS

        self.scheduler.add_job(
            func=self._renew_token,
            trigger=IntervalTrigger(
                # add seconds to current time to make sure that the first refresh
                start_date=datetime.now()
                + timedelta(seconds=remaining_life_in_seconds),
                seconds=refresh_interval_in_seconds,
            ),
        )


class VaultSecretManagerConfig(SecretManagerConfig):
    """Configuration for the Vault secret manager."""

    def __init__(
        self,
        url: Text,
        token: Text,
        secrets_path: Text,
        transit_mount_point: Text = "transit",
        namespace: Optional[Text] = None,
        mount_point: Optional[Text] = None,
    ) -> None:
        """Initialise the VaultSecretManagerConfig.

        Args:
            url: The URL of the vault server.
            token: The token to authenticate with the vault server.
            secrets_path: The path to the secrets in the vault server.
            transit_mount_point: The mount point of the transit engine.
            namespace: The namespace in which secrets reside in.
        """
        super().__init__(VAULT_SECRET_MANAGER_NAME)
        self.url = url
        self.token = token
        self.secrets_path = secrets_path
        self.transit_mount_point = transit_mount_point
        self.namespace = namespace
        self.mount_point = mount_point


@dataclass
class VaultSecretManagerNonStrictConfig:
    """Non-Strict configuration for the Vault secret manager.

    It is used to validate and merge configurations
    from environment variables and endpoints.yml file.
    """

    url: Optional[Text]
    token: Optional[Text]
    secrets_path: Optional[Text]
    transit_mount_point: Optional[Text]
    namespace: Optional[Text] = None
    mount_point: Optional[Text] = None

    def is_empty(self) -> bool:
        """Check if all the values are empty."""
        return (
            (self.url is None or self.url == "")
            and (self.token is None or self.token == "")
            and (self.secrets_path is None or self.secrets_path == "")
            and (self.transit_mount_point is None or self.transit_mount_point == "")
            and (self.namespace is None or self.namespace == "")
            and (self.mount_point is None or self.mount_point == "")
        )

    def is_valid(self) -> bool:
        """Check if all the values are valid.

        url and token are required,
        secrets_path is required,
        transit_mount_point is optional, but if provided, it must not be empty.

        Returns:
            True if all the values are valid, False otherwise.
        """
        return (
            self.url is not None
            and self.url != ""
            and self.token is not None
            and self.token != ""
            and self.secrets_path is not None
            and self.secrets_path != ""
            and self._is_optional_value_valid(self.transit_mount_point)
            and self._is_optional_value_valid(self.namespace)
            and self._is_optional_value_valid(self.mount_point)
        )

    @staticmethod
    def _is_optional_value_valid(value: Optional[Text]) -> bool:
        """Check if the optional value is valid.

        Args:
            value: The optional value to check.

        Returns:
            True if the optional value is valid, False otherwise.
        """
        return value is None or value != ""

    def merge(
        self, other: "VaultSecretManagerNonStrictConfig"
    ) -> "VaultSecretManagerNonStrictConfig":
        """Merge two VaultSecretManagerEnvConfig objects.

        Args:
            other: The other VaultSecretManagerEnvConfig object.

        Returns:
            The merged VaultSecretManagerEnvConfig object.
        """
        return VaultSecretManagerNonStrictConfig(
            url=self.url or other.url,
            token=self.token or other.token,
            secrets_path=self.secrets_path or other.secrets_path,
            transit_mount_point=self.transit_mount_point or other.transit_mount_point,
            namespace=self.namespace or other.namespace,
            mount_point=self.mount_point or other.mount_point,
        )


def get_vault_endpoint_reader(
    endpoint_trait: EndpointTrait,
) -> Optional[VaultEndpointConfigReader]:
    """Return VaultEndpointConfigReader for a given endpoint type.

    Args:
        endpoint_trait: The endpoint trait to create the reader for.

    Returns:
        The reader associated with the endpoint's type.
    """
    if (
        endpoint_trait.endpoint_type == TRACKER_STORE_ENDPOINT_TYPE
        and TrackerStoreEndpointValidator(
            endpoint_reader=EndpointReader(
                endpoint_config=endpoint_trait.endpoint_config
            )
        ).is_endpoint_config_valid()
    ):
        return VaultEndpointConfigReader(endpoint_config=endpoint_trait.endpoint_config)

    return None



from __future__ import annotations

from copy import deepcopy
from dataclasses import asdict, dataclass, field
from typing import (
    Any,
    Dict,
    Optional,
    Set,
)

import structlog
from pydantic import BaseModel

from rasa.shared.constants import (
    API_BASE_CONFIG_KEY,
    API_KEY,
    API_TYPE_CONFIG_KEY,
    API_VERSION_CONFIG_KEY,
    AZURE_API_TYPE,
    AZURE_OPENAI_PROVIDER,
    DEPLOYMENT_CONFIG_KEY,
    DEPLOYMENT_NAME_CONFIG_KEY,
    ENGINE_CONFIG_KEY,
    LANGCHAIN_TYPE_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODEL_NAME_CONFIG_KEY,
    N_REPHRASES_CONFIG_KEY,
    OPENAI_API_BASE_CONFIG_KEY,
    OPENAI_API_TYPE_CONFIG_KEY,
    OPENAI_API_VERSION_CONFIG_KEY,
    PROVIDER_CONFIG_KEY,
    RASA_TYPE_CONFIG_KEY,
    REQUEST_TIMEOUT_CONFIG_KEY,
    STREAM_CONFIG_KEY,
    TIMEOUT_CONFIG_KEY,
)
from rasa.shared.providers._configs.azure_entra_id_config import (
    AzureEntraIDOAuthConfig,
    AzureEntraIDOAuthType,
)
from rasa.shared.providers._configs.oauth_config import (
    OAUTH_KEY,
    OAUTH_TYPE_FIELD,
    OAuth,
)
from rasa.shared.providers._configs.utils import (
    raise_deprecation_warnings,
    resolve_aliases,
    validate_forbidden_keys,
    validate_required_keys,
)
from rasa.shared.utils.common import class_from_module_path

structlogger = structlog.get_logger()

DEPRECATED_ALIASES_TO_STANDARD_KEY_MAPPING = {
    # Deployment name aliases
    DEPLOYMENT_NAME_CONFIG_KEY: DEPLOYMENT_CONFIG_KEY,
    ENGINE_CONFIG_KEY: DEPLOYMENT_CONFIG_KEY,
    # Provider aliases
    RASA_TYPE_CONFIG_KEY: PROVIDER_CONFIG_KEY,
    LANGCHAIN_TYPE_CONFIG_KEY: PROVIDER_CONFIG_KEY,
    # API type aliases
    OPENAI_API_TYPE_CONFIG_KEY: API_TYPE_CONFIG_KEY,
    # API base aliases
    OPENAI_API_BASE_CONFIG_KEY: API_BASE_CONFIG_KEY,
    # API version aliases
    OPENAI_API_VERSION_CONFIG_KEY: API_VERSION_CONFIG_KEY,
    # Model name aliases
    MODEL_NAME_CONFIG_KEY: MODEL_CONFIG_KEY,
    # Timeout aliases
    REQUEST_TIMEOUT_CONFIG_KEY: TIMEOUT_CONFIG_KEY,
}

REQUIRED_KEYS = [DEPLOYMENT_CONFIG_KEY]

FORBIDDEN_KEYS = [
    STREAM_CONFIG_KEY,
    N_REPHRASES_CONFIG_KEY,
]


class OAuthConfigWrapper(OAuth, BaseModel):
    """Wrapper for OAuth configuration.

    It's main purpose is to provide to_dict method which is used to serialize
    the oauth configuration to the original format.

    """

    # Pydantic configuration to allow arbitrary user defined types
    class Config:
        arbitrary_types_allowed = True

    oauth: OAuth
    original_config: Dict[str, Any]

    def get_bearer_token(self) -> str:
        """Returns a bearer token."""
        return self.oauth.get_bearer_token()

    def to_dict(self) -> Dict[str, Any]:
        """Converts the OAuth configuration to the original format."""
        return self.original_config

    @staticmethod
    def _valid_type_values() -> Set[str]:
        """Returns the valid built-in values for the `type` field in the `oauth`."""
        return AzureEntraIDOAuthType.valid_string_values()

    @classmethod
    def from_dict(cls, oauth_config: Dict[str, Any]) -> OAuthConfigWrapper:
        """Initializes a dataclass from the passed config.

        Args:
            oauth_config: (dict) The config from which to initialize.

        Returns:
            AzureOAuthConfig
        """
        original_config = deepcopy(oauth_config)

        oauth_type: Optional[str] = oauth_config.get(OAUTH_TYPE_FIELD, None)

        if oauth_type is None:
            message = (
                "Oauth configuration must contain "
                f"'{OAUTH_TYPE_FIELD}' field and it must be set to one of the "
                f"following values: {OAuthConfigWrapper._valid_type_values()}, "
                f"or to the path of module which is "
                f"implementing {OAuth.__name__} protocol."
            )
            structlogger.error(
                "azure_oauth_config.missing_oauth_type",
                message=message,
            )
            raise ValueError(message)

        if oauth_type in AzureEntraIDOAuthType.valid_string_values():
            return cls(
                oauth=AzureEntraIDOAuthConfig.from_dict(oauth_config),
                original_config=original_config,
            )

        module = class_from_module_path(oauth_type)

        if not issubclass(module, OAuth):
            message = (
                f"Module {oauth_type} does not implement "
                f"{OAuth.__name__} interface."
            )
            structlogger.error(
                "azure_oauth_config.invalid_oauth_module",
                message=message,
            )
            raise ValueError(message)

        return cls(
            oauth=module.from_dict(oauth_config), original_config=original_config
        )


@dataclass
class AzureOpenAIClientConfig:
    """Parses configuration for Azure OpenAI client, resolves aliases and
    raises deprecation warnings.

    Raises:
        ValueError: Raised in cases of invalid configuration:
            - If any of the required configuration keys are missing.
            - If `api_type` has a value different from `azure`.
    """

    deployment: str

    model: Optional[str]
    api_base: Optional[str]
    api_version: Optional[str]
    # API Type is not used by LiteLLM backend, but we define
    # it here for backward compatibility.
    api_type: Optional[str] = AZURE_API_TYPE
    # Provider is not used by LiteLLM backend, but we define it here since it's
    # used as switch between different clients.
    provider: str = AZURE_OPENAI_PROVIDER

    # OAuth related parameters
    oauth: Optional[OAuthConfigWrapper] = None

    extra_parameters: dict = field(default_factory=dict)

    def __post_init__(self) -> None:
        if self.provider != AZURE_OPENAI_PROVIDER:
            message = f"Provider must be set to '{AZURE_OPENAI_PROVIDER}'."
            structlogger.error(
                "azure_openai_client_config.validation_error",
                message=message,
                provider=self.provider,
            )
            raise ValueError(message)
        if self.deployment is None:
            message = "Deployment cannot be set to None."
            structlogger.error(
                "azure_openai_client_config.validation_error",
                message=message,
                deployment=self.deployment,
            )
            raise ValueError(message)

    @classmethod
    def from_dict(cls, config: dict) -> AzureOpenAIClientConfig:
        """Initializes a dataclass from the passed config.

        Args:
            config: (dict) The config from which to initialize.

        Raises:
            ValueError: Raised in cases of invalid configuration:
                - If any of the required configuration keys are missing.
                - If `api_type` has a value different from `azure`.

        Returns:
            AzureOpenAIClientConfig
        """
        # Check for deprecated keys
        raise_deprecation_warnings(config, DEPRECATED_ALIASES_TO_STANDARD_KEY_MAPPING)
        # Resolve any potential aliases
        config = cls.resolve_config_aliases(config)
        # Validate that required keys are set
        validate_required_keys(config, REQUIRED_KEYS)
        # Validate that the forbidden keys are not present
        validate_forbidden_keys(config, FORBIDDEN_KEYS)
        # Init client config

        cls._validate_authentication_configuration(config)

        has_oauth_key = config.get(OAUTH_KEY, None) is not None
        oauth = (
            OAuthConfigWrapper.from_dict(config.pop(OAUTH_KEY))
            if has_oauth_key
            else None
        )

        this = AzureOpenAIClientConfig(
            # Required parameters
            deployment=config.pop(DEPLOYMENT_CONFIG_KEY),
            # Pop the 'provider' key. Currently, it's *optional* because of
            # backward compatibility with older versions.
            provider=config.pop(PROVIDER_CONFIG_KEY, AZURE_OPENAI_PROVIDER),
            # Optional
            api_type=config.pop(API_TYPE_CONFIG_KEY, AZURE_API_TYPE),
            model=config.pop(MODEL_CONFIG_KEY, None),
            # Optional, can also be set through environment variables
            # in clients.
            api_base=config.pop(API_BASE_CONFIG_KEY, None),
            api_version=config.pop(API_VERSION_CONFIG_KEY, None),
            # OAuth related parameters, set only if auth_type is set to 'entra_id'
            oauth=oauth,
            # The rest of parameters (e.g. model parameters) are considered
            # as extra parameters (this also includes timeout).
            extra_parameters=config,
        )
        return this

    def to_dict(self) -> dict:
        """Converts the config instance into a dictionary."""
        d = asdict(self)
        # Extra parameters should also be on the top level
        d.pop("extra_parameters", None)
        d.update(self.extra_parameters)

        d.pop("oauth", None)
        d.update({"oauth": self.oauth.to_dict()} if self.oauth else {})
        return d

    @staticmethod
    def resolve_config_aliases(config: Dict[str, Any]) -> Dict[str, Any]:
        return resolve_aliases(config, DEPRECATED_ALIASES_TO_STANDARD_KEY_MAPPING)

    @staticmethod
    def _validate_authentication_configuration(config: Dict[str, Any]) -> None:
        """Validates the authentication configuration."""
        has_api_key = config.get(API_KEY, None) is not None
        has_oauth_key = config.get(OAUTH_KEY, None) is not None

        if has_api_key and has_oauth_key:
            message = (
                "Azure OpenAI client configuration cannot contain "
                f"both '{API_KEY}' and '{OAUTH_KEY}' fields. Please provide either "
                f"'{API_KEY}' or '{OAUTH_KEY}' fields."
            )
            structlogger.error(
                "azure_openai_client_config.multiple_auth_types_specified",
                message=message,
            )
            raise ValueError(message)


def is_azure_openai_config(config: dict) -> bool:
    """Check whether the configuration is meant to configure
    an Azure OpenAI client.
    """
    # Resolve any aliases that are specific to Azure OpenAI configuration
    config = AzureOpenAIClientConfig.resolve_config_aliases(config)

    # Case: Configuration contains `provider: azure`.
    if config.get(PROVIDER_CONFIG_KEY) == AZURE_OPENAI_PROVIDER:
        return True

    # Case: Configuration contains `deployment` key
    # (specific to Azure OpenAI configuration)
    if (
        config.get(DEPLOYMENT_CONFIG_KEY) is not None
        and config.get(PROVIDER_CONFIG_KEY) is None
    ):
        return True

    return False



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any, Dict, List

from rasa.dialogue_understanding.commands.command import Command
from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxManager,
    CommandSyntaxVersion,
)
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker


@dataclass
class ChangeFlowCommand(Command):
    """A command to indicate a change of flows was requested by the command
    generator.
    """

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "change_flow"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> ChangeFlowCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        return ChangeFlowCommand()

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        # the change flow command is not actually pushing anything to the tracker,
        # but it is predicted by the MultiStepLLMCommandGenerator and used internally
        return []

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, ChangeFlowCommand)

    def __hash__(self) -> int:
        return hash(self.command())

    def to_dsl(self) -> str:
        """Converts the command to a DSL string."""
        mapper = {
            CommandSyntaxVersion.v1: "ChangeFlow()",
            CommandSyntaxVersion.v2: "change",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )

    @staticmethod
    def from_dsl(match: re.Match, **kwargs: Any) -> ChangeFlowCommand:
        """Converts the DSL string to a command."""
        return ChangeFlowCommand()

    @staticmethod
    def regex_pattern() -> str:
        mapper = {
            CommandSyntaxVersion.v1: r"ChangeFlow\(\)",
            CommandSyntaxVersion.v2: r"^[^\w]*change$",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Text

import structlog

from rasa.core.actions import action
from rasa.core.channels import OutputChannel
from rasa.core.nlg import NaturalLanguageGenerator
from rasa.dialogue_understanding.patterns.collect_information import (
    CollectInformationPatternFlowStackFrame,
)
from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack
from rasa.dialogue_understanding.stack.frames import (
    BaseFlowStackFrame,
    PatternFlowStackFrame,
)
from rasa.dialogue_understanding.stack.frames.dialogue_stack_frame import (
    DialogueStackFrame,
)
from rasa.dialogue_understanding.stack.frames.flow_stack_frame import (
    FlowStackFrameType,
    UserFlowStackFrame,
)
from rasa.dialogue_understanding.stack.utils import top_user_flow_frame
from rasa.shared.constants import RASA_DEFAULT_FLOW_PATTERN_PREFIX
from rasa.shared.core.constants import ACTION_CORRECT_FLOW_SLOT
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import Event, SlotSet
from rasa.shared.core.flows.flow import END_STEP, START_STEP, ContinueFlowStep
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()

FLOW_PATTERN_CORRECTION_ID = RASA_DEFAULT_FLOW_PATTERN_PREFIX + "correction"


@dataclass
class CorrectionPatternFlowStackFrame(PatternFlowStackFrame):
    """A pattern flow stack frame which gets added if a slot value is corrected."""

    flow_id: str = FLOW_PATTERN_CORRECTION_ID
    """The ID of the flow."""
    is_reset_only: bool = False
    """Whether the correction is only a reset of the flow.

    This is the case if all corrected slots have `ask_before_filling=True`.
    In this case, we do not set their value directly but rather reset the flow
    to the position where the first question is asked to fill the slot."""
    corrected_slots: Dict[str, Any] = field(default_factory=dict)
    """The slots that were corrected."""
    reset_flow_id: Optional[str] = None
    """The ID of the flow to reset to."""
    reset_step_id: Optional[str] = None
    """The ID of the step to reset to."""
    new_slot_values: List[Any] = field(default_factory=list)
    """The new values for the corrected slots."""

    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        return FLOW_PATTERN_CORRECTION_ID

    @staticmethod
    def from_dict(data: Dict[Text, Any]) -> CorrectionPatternFlowStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        new_slot_values = [
            val.get("value") for _, val in data["corrected_slots"].items()
        ]

        return CorrectionPatternFlowStackFrame(
            frame_id=data["frame_id"],
            step_id=data["step_id"],
            is_reset_only=data["is_reset_only"],
            corrected_slots=data["corrected_slots"],
            reset_flow_id=data["reset_flow_id"],
            reset_step_id=data["reset_step_id"],
            new_slot_values=new_slot_values,
        )


class ActionCorrectFlowSlot(action.Action):
    """Action which corrects a slots value in a flow."""

    def __init__(self) -> None:
        """Creates a `ActionCorrectFlowSlot`."""
        super().__init__()

    def name(self) -> Text:
        """Return the flow name."""
        return ACTION_CORRECT_FLOW_SLOT

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Correct the slots."""
        stack = tracker.stack
        if not (top := stack.top()):
            structlogger.warning("action.correct_flow_slot.no_active_flow")
            return []

        if not isinstance(top, CorrectionPatternFlowStackFrame):
            structlogger.warning(
                "action.correct_flow_slot.no_correction_frame", top=top
            )
            return []

        events: List[Event] = []
        if top.reset_flow_id:
            updated_stack = reset_stack_on_tracker_to_prior_state(
                top.reset_flow_id, top.reset_step_id or START_STEP, tracker
            )
            events.extend(tracker.create_stack_updated_events(updated_stack))

        events.extend(
            [
                SlotSet(name, value=val.get("value"), filled_by=val.get("filled_by"))
                for name, val in top.corrected_slots.items()
            ]
        )
        return events


def find_previous_state_to_reset_to(
    reset_flow_id: str, reset_step_id: str, tracker: DialogueStateTracker
) -> Optional[DialogueStack]:
    """Find the previous stack state where the reset step was the active stack frame.

    Looks for the most recent state of the stack in the tracker where the top
    flow has the same flow id and step id as the reset flow id and reset step id.
    """
    stack_to_reset_to: Optional[DialogueStack] = None
    for previous_stack_state in tracker.previous_stack_states():
        top_frame = previous_stack_state.top()
        if (
            isinstance(top_frame, BaseFlowStackFrame)
            and top_frame.flow_id == reset_flow_id
            and top_frame.step_id == reset_step_id
        ):
            stack_to_reset_to = previous_stack_state
    return stack_to_reset_to


def slice_of_stack_below_target(
    target_frame: DialogueStackFrame, stack: DialogueStack
) -> List[DialogueStackFrame]:
    """Get all the frames from the current stack up until the target frame.

    The target frame is not included."""
    replacement_stack: List[DialogueStackFrame] = []

    for frame in stack.frames:
        if frame.frame_id == target_frame.frame_id:
            break
        replacement_stack.append(frame)
    return replacement_stack


def slice_of_stack_with_target_and_above(
    target_frame: DialogueStackFrame, stack: DialogueStack
) -> List[DialogueStackFrame]:
    """Get all the frames from the current stack starting at the target frame."""
    replacement_frames: List[DialogueStackFrame] = []
    for frame in reversed(stack.frames):
        replacement_frames.insert(0, frame)  # push to front, since reversed
        if frame.frame_id == target_frame.frame_id:
            break
    return replacement_frames


def set_topmost_flow_frame_to_continue(stack_frames: List[DialogueStackFrame]) -> None:
    """Ensure that the topmost flow frame continues its step.

    This ensures that the topmost step can be re-executed."""
    for frame in reversed(stack_frames):
        if isinstance(frame, BaseFlowStackFrame) and frame.step_id != START_STEP:
            frame.step_id = ContinueFlowStep.continue_step_for_id(frame.step_id)
            break


def create_termination_frames_for_missing_frames(
    new_stack_frames: List[DialogueStackFrame], previous_stack: DialogueStack
) -> List[DialogueStackFrame]:
    """Terminate frames that are part in the previous stack but not in the new stack.

    The frames are terminated by setting them to the END_STEP. This allows
    them to be properly wrapped up.
    """

    reused_frame_ids = {frame.frame_id for frame in new_stack_frames}

    frames_to_terminate: List[DialogueStackFrame] = []
    for frame in reversed(previous_stack.frames):
        if frame.frame_id in reused_frame_ids:
            # this frame already exists in the replacement stack, skip it
            # shouldn't be terminated
            continue
        if (
            isinstance(frame, UserFlowStackFrame)
            and frame.frame_type == FlowStackFrameType.CALL
        ):
            frame.step_id = ContinueFlowStep.continue_step_for_id(END_STEP)
        if isinstance(frame, CollectInformationPatternFlowStackFrame):
            frame.step_id = ContinueFlowStep.continue_step_for_id(END_STEP)
        frames_to_terminate.insert(0, frame)
    return frames_to_terminate


def reset_stack_on_tracker_to_prior_state(
    reset_flow_id: str, reset_step_id: str, tracker: DialogueStateTracker
) -> DialogueStack:
    """Reset the stack on the tracker to the prior state."""
    # assumption is that we have already been at the reset step id in the reset flow
    # at some point before in the tracker. we'll search for that point in time and
    # update the current stack to be as close to that time as possible,
    # jumping back in time.
    stack_to_reset_to: Optional[DialogueStack] = find_previous_state_to_reset_to(
        reset_flow_id, reset_step_id, tracker
    )

    if not stack_to_reset_to:
        structlogger.warning(
            "action.correct_flow_slot.no_frame_found",
            reset_step_id=reset_step_id,
            reset_flow_id=reset_flow_id,
        )
        return tracker.stack

    target_frame = top_user_flow_frame(stack_to_reset_to)

    if not target_frame:
        structlogger.warning(
            "action.correct_flow_slot.no_target_frame_found",
            reset_step_id=reset_step_id,
            reset_flow_id=reset_flow_id,
            stack_to_reset_to=stack_to_reset_to,
        )
        return tracker.stack

    current_stack = tracker.stack

    # Approach we use to reset the stack:
    # 1. Search through all past stack states and identify the most recent
    #    state where the stack had the given flow and step id on top. This
    #    stack is the reference stack.
    # 2. On the reference stack, identify the top most user flow frame that
    #    was not started through a flow call. This frame is the reference frame.
    # 3. Create a new stack by stacking the following frames: A) all frames
    #    from the current stack that are strictly below the reference
    #    frame. B) all frames including the reference frame and frames above
    #    it from the reference stack. C) all frames from the current stack
    #    that are strictly above the reference frame and set them to their
    #    end step.
    # 4. This will ensure A) we keep anything from the current stack that is
    #    below the current active user flow. B) We reset the current user
    #    flow to the desired position. C) we wrap up all unfinished frames
    #    above the current user flow from the current state.

    replacement_stack: List[DialogueStackFrame] = []

    # get all the frames from the current stack up until the target frame
    # and put them into the replacement stack. the target frame is not included
    replacement_stack.extend(slice_of_stack_below_target(target_frame, current_stack))

    replacement_stack.extend(
        slice_of_stack_with_target_and_above(target_frame, stack_to_reset_to)
    )

    # ensure that we continue the topmost frames step
    set_topmost_flow_frame_to_continue(replacement_stack)

    # terminate all the frames from the original stack that are not part of the
    # replacement stack. this will add these frames to the replacement stack,
    # but with their step set to END_STEP. this allows them to be properly
    # wrapped up before we head into the correction.
    replacement_stack.extend(
        create_termination_frames_for_missing_frames(replacement_stack, current_stack)
    )
    return DialogueStack(frames=replacement_stack)



from __future__ import annotations

import logging
import typing
from collections import OrderedDict
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Text, Tuple, Type

import numpy as np

import rasa.nlu.utils.bilou_utils as bilou_utils
import rasa.shared.utils.io
import rasa.utils.train_utils
from rasa.engine.graph import ExecutionContext, GraphComponent
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.nlu.constants import TOKENS_NAMES
from rasa.nlu.extractors.extractor import EntityExtractorMixin
from rasa.nlu.test import determine_token_labels
from rasa.nlu.tokenizers.spacy_tokenizer import POS_TAG_KEY
from rasa.nlu.tokenizers.tokenizer import Token, Tokenizer
from rasa.shared.constants import DOCS_URL_COMPONENTS
from rasa.shared.nlu.constants import (
    ENTITIES,
    ENTITY_ATTRIBUTE_GROUP,
    ENTITY_ATTRIBUTE_ROLE,
    ENTITY_ATTRIBUTE_TYPE,
    NO_ENTITY_TAG,
    SPLIT_ENTITIES_BY_COMMA,
    SPLIT_ENTITIES_BY_COMMA_DEFAULT_VALUE,
    TEXT,
)
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.utils.tensorflow.constants import BILOU_FLAG, FEATURIZERS

logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from sklearn_crfsuite import CRF


CONFIG_FEATURES = "features"


class CRFToken:
    def __init__(
        self,
        text: Text,
        pos_tag: Text,
        pattern: Dict[Text, Any],
        dense_features: np.ndarray,
        entity_tag: Text,
        entity_role_tag: Text,
        entity_group_tag: Text,
    ):
        self.text = text
        self.pos_tag = pos_tag
        self.pattern = pattern
        self.dense_features = dense_features
        self.entity_tag = entity_tag
        self.entity_role_tag = entity_role_tag
        self.entity_group_tag = entity_group_tag

    def to_dict(self) -> Dict[str, Any]:
        return {
            "text": self.text,
            "pos_tag": self.pos_tag,
            "pattern": self.pattern,
            "dense_features": [str(x) for x in list(self.dense_features)],
            "entity_tag": self.entity_tag,
            "entity_role_tag": self.entity_role_tag,
            "entity_group_tag": self.entity_group_tag,
        }

    @classmethod
    def create_from_dict(cls, data: Dict[str, Any]) -> "CRFToken":
        return cls(
            data["text"],
            data["pos_tag"],
            data["pattern"],
            np.array([float(x) for x in data["dense_features"]]),
            data["entity_tag"],
            data["entity_role_tag"],
            data["entity_group_tag"],
        )


class CRFEntityExtractorOptions(str, Enum):
    """Features that can be used for the 'CRFEntityExtractor'."""

    PATTERN = "pattern"
    LOW = "low"
    TITLE = "title"
    PREFIX5 = "prefix5"
    PREFIX2 = "prefix2"
    SUFFIX5 = "suffix5"
    SUFFIX3 = "suffix3"
    SUFFIX2 = "suffix2"
    SUFFIX1 = "suffix1"
    BIAS = "bias"
    POS = "pos"
    POS2 = "pos2"
    UPPER = "upper"
    DIGIT = "digit"
    TEXT_DENSE_FEATURES = "text_dense_features"
    ENTITY = "entity"


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.ENTITY_EXTRACTOR, is_trainable=True
)
class CRFEntityExtractor(GraphComponent, EntityExtractorMixin):
    """Implements conditional random fields (CRF) to do named entity recognition."""

    function_dict: Dict[Text, Callable[[CRFToken], Any]] = {  # noqa: RUF012
        CRFEntityExtractorOptions.LOW: lambda crf_token: crf_token.text.lower(),
        CRFEntityExtractorOptions.TITLE: lambda crf_token: crf_token.text.istitle(),
        CRFEntityExtractorOptions.PREFIX5: lambda crf_token: crf_token.text[:5],
        CRFEntityExtractorOptions.PREFIX2: lambda crf_token: crf_token.text[:2],
        CRFEntityExtractorOptions.SUFFIX5: lambda crf_token: crf_token.text[-5:],
        CRFEntityExtractorOptions.SUFFIX3: lambda crf_token: crf_token.text[-3:],
        CRFEntityExtractorOptions.SUFFIX2: lambda crf_token: crf_token.text[-2:],
        CRFEntityExtractorOptions.SUFFIX1: lambda crf_token: crf_token.text[-1:],
        CRFEntityExtractorOptions.BIAS: lambda _: "bias",
        CRFEntityExtractorOptions.POS: lambda crf_token: crf_token.pos_tag,
        CRFEntityExtractorOptions.POS2: lambda crf_token: crf_token.pos_tag[:2]
        if crf_token.pos_tag is not None
        else None,
        CRFEntityExtractorOptions.UPPER: lambda crf_token: crf_token.text.isupper(),
        CRFEntityExtractorOptions.DIGIT: lambda crf_token: crf_token.text.isdigit(),
        CRFEntityExtractorOptions.PATTERN: lambda crf_token: crf_token.pattern,
        CRFEntityExtractorOptions.TEXT_DENSE_FEATURES: (
            lambda crf_token: CRFEntityExtractor._convert_dense_features_for_crfsuite(
                crf_token
            )
        ),
        CRFEntityExtractorOptions.ENTITY: lambda crf_token: crf_token.entity_tag,
    }

    @classmethod
    def required_components(cls) -> List[Type]:
        """Components that should be included in the pipeline before this component."""
        return [Tokenizer]

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """The component's default config (see parent class for full docstring)."""
        return {
            # BILOU_flag determines whether to use BILOU tagging or not.
            # More rigorous however requires more examples per entity
            # rule of thumb: use only if more than 100 egs. per entity
            BILOU_FLAG: True,
            # Split entities by comma, this makes sense e.g. for a list of ingredients
            # in a recipie, but it doesn't make sense for the parts of an address
            SPLIT_ENTITIES_BY_COMMA: True,
            # crf_features is [before, token, after] array with before, token,
            # after holding keys about which features to use for each token,
            # for example, 'title' in array before will have the feature
            # "is the preceding token in title case?"
            # POS features require SpacyTokenizer
            # pattern feature require RegexFeaturizer
            CONFIG_FEATURES: [
                [
                    CRFEntityExtractorOptions.LOW,
                    CRFEntityExtractorOptions.TITLE,
                    CRFEntityExtractorOptions.UPPER,
                ],
                [
                    CRFEntityExtractorOptions.LOW,
                    CRFEntityExtractorOptions.BIAS,
                    CRFEntityExtractorOptions.PREFIX5,
                    CRFEntityExtractorOptions.PREFIX2,
                    CRFEntityExtractorOptions.SUFFIX5,
                    CRFEntityExtractorOptions.SUFFIX3,
                    CRFEntityExtractorOptions.SUFFIX2,
                    CRFEntityExtractorOptions.UPPER,
                    CRFEntityExtractorOptions.TITLE,
                    CRFEntityExtractorOptions.DIGIT,
                    CRFEntityExtractorOptions.PATTERN,
                ],
                [
                    CRFEntityExtractorOptions.LOW,
                    CRFEntityExtractorOptions.TITLE,
                    CRFEntityExtractorOptions.UPPER,
                ],
            ],
            # The maximum number of iterations for optimization algorithms.
            "max_iterations": 50,
            # weight of the L1 regularization
            "L1_c": 0.1,
            # weight of the L2 regularization
            "L2_c": 0.1,
            # Name of dense featurizers to use.
            # If list is empty all available dense features are used.
            "featurizers": [],
        }

    def __init__(
        self,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        entity_taggers: Optional[Dict[Text, "CRF"]] = None,
    ) -> None:
        """Creates an instance of entity extractor."""
        self.component_config = config
        self._model_storage = model_storage
        self._resource = resource

        self.entity_taggers = entity_taggers

        self.crf_order = [
            ENTITY_ATTRIBUTE_TYPE,
            ENTITY_ATTRIBUTE_ROLE,
            ENTITY_ATTRIBUTE_GROUP,
        ]

        self._validate_configuration()

        self.split_entities_config = rasa.utils.train_utils.init_split_entities(
            config[SPLIT_ENTITIES_BY_COMMA], SPLIT_ENTITIES_BY_COMMA_DEFAULT_VALUE
        )

    def _validate_configuration(self) -> None:
        if len(self.component_config.get(CONFIG_FEATURES, [])) % 2 != 1:
            raise ValueError(
                "Need an odd number of crf feature lists to have a center word."
            )

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> CRFEntityExtractor:
        """Creates a new untrained component (see parent class for full docstring)."""
        return cls(config, model_storage, resource)

    @staticmethod
    def required_packages() -> List[Text]:
        """Any extra python dependencies required for this component to run."""
        return ["sklearn_crfsuite", "sklearn"]

    def train(self, training_data: TrainingData) -> Resource:
        """Trains the extractor on a data set."""
        # checks whether there is at least one
        # example with an entity annotation
        if not training_data.entity_examples:
            logger.debug(
                "No training examples with entities present. Skip training"
                "of 'CRFEntityExtractor'."
            )
            return self._resource

        self.check_correct_entity_annotations(training_data)

        if self.component_config[BILOU_FLAG]:
            bilou_utils.apply_bilou_schema(training_data)

        # only keep the CRFs for tags we actually have training data for
        self._update_crf_order(training_data)

        # filter out pre-trained entity examples
        entity_examples = self.filter_trainable_entities(training_data.nlu_examples)
        entity_examples = [
            message
            for message in entity_examples
            if message.features_present(
                attribute=TEXT, featurizers=self.component_config.get(FEATURIZERS)
            )
        ]
        dataset = [self._convert_to_crf_tokens(example) for example in entity_examples]

        self.entity_taggers = self.train_model(
            dataset, self.component_config, self.crf_order
        )

        self.persist(dataset)

        return self._resource

    def _update_crf_order(self, training_data: TrainingData) -> None:
        """Train only CRFs we actually have training data for."""
        _crf_order = []

        for tag_name in self.crf_order:
            if tag_name == ENTITY_ATTRIBUTE_TYPE and training_data.entities:
                _crf_order.append(ENTITY_ATTRIBUTE_TYPE)
            elif tag_name == ENTITY_ATTRIBUTE_ROLE and training_data.entity_roles:
                _crf_order.append(ENTITY_ATTRIBUTE_ROLE)
            elif tag_name == ENTITY_ATTRIBUTE_GROUP and training_data.entity_groups:
                _crf_order.append(ENTITY_ATTRIBUTE_GROUP)

        self.crf_order = _crf_order

    def process(self, messages: List[Message]) -> List[Message]:
        """Augments messages with entities."""
        for message in messages:
            entities = self.extract_entities(message)
            entities = self.add_extractor_name(entities)
            message.set(
                ENTITIES, message.get(ENTITIES, []) + entities, add_to_output=True
            )

        return messages

    def extract_entities(self, message: Message) -> List[Dict[Text, Any]]:
        """Extract entities from the given message using the trained model(s)."""
        if self.entity_taggers is None or not message.features_present(
            attribute=TEXT, featurizers=self.component_config.get(FEATURIZERS)
        ):
            return []

        tokens = message.get(TOKENS_NAMES[TEXT])
        crf_tokens = self._convert_to_crf_tokens(message)

        predictions: Dict[Text, List[Dict[Text, float]]] = {}
        for tag_name, entity_tagger in self.entity_taggers.items():
            # use predicted entity tags as features for second level CRFs
            include_tag_features = tag_name != ENTITY_ATTRIBUTE_TYPE
            if include_tag_features:
                self._add_tag_to_crf_token(crf_tokens, predictions)

            features = self._crf_tokens_to_features(
                crf_tokens, self.component_config, include_tag_features
            )
            predictions[tag_name] = entity_tagger.predict_marginals_single(features)

        # convert predictions into a list of tags and a list of confidences
        tags, confidences = self._tag_confidences(tokens, predictions)

        return self.convert_predictions_into_entities(
            message.get(TEXT), tokens, tags, self.split_entities_config, confidences
        )

    def _add_tag_to_crf_token(
        self,
        crf_tokens: List[CRFToken],
        predictions: Dict[Text, List[Dict[Text, float]]],
    ) -> None:
        """Add predicted entity tags to CRF tokens."""
        if ENTITY_ATTRIBUTE_TYPE in predictions:
            _tags, _ = self._most_likely_tag(predictions[ENTITY_ATTRIBUTE_TYPE])
            for tag, token in zip(_tags, crf_tokens):
                token.entity_tag = tag

    def _most_likely_tag(
        self, predictions: List[Dict[Text, float]]
    ) -> Tuple[List[Text], List[float]]:
        """Get the entity tags with the highest confidence.

        Args:
            predictions: list of mappings from entity tag to confidence value

        Returns:
            List of entity tags and list of confidence values.
        """
        _tags = []
        _confidences = []

        for token_predictions in predictions:
            tag = max(token_predictions, key=lambda key: token_predictions[key])
            _tags.append(tag)

            if self.component_config[BILOU_FLAG]:
                # if we are using BILOU flags, we will sum up the prob
                # of the B, I, L and U tags for an entity
                _confidences.append(
                    sum(
                        _confidence
                        for _tag, _confidence in token_predictions.items()
                        if bilou_utils.tag_without_prefix(tag)
                        == bilou_utils.tag_without_prefix(_tag)
                    )
                )
            else:
                _confidences.append(token_predictions[tag])

        return _tags, _confidences

    def _tag_confidences(
        self, tokens: List[Token], predictions: Dict[Text, List[Dict[Text, float]]]
    ) -> Tuple[Dict[Text, List[Text]], Dict[Text, List[float]]]:
        """Get most likely tag predictions with confidence values for tokens."""
        tags = {}
        confidences = {}

        for tag_name, predicted_tags in predictions.items():
            if len(tokens) != len(predicted_tags):
                raise Exception(
                    "Inconsistency in amount of tokens between crfsuite and message"
                )

            _tags, _confidences = self._most_likely_tag(predicted_tags)

            if self.component_config[BILOU_FLAG]:
                _tags, _confidences = bilou_utils.ensure_consistent_bilou_tagging(
                    _tags, _confidences
                )

            confidences[tag_name] = _confidences
            tags[tag_name] = _tags

        return tags, confidences

    @classmethod
    def load(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> CRFEntityExtractor:
        """Loads trained component (see parent class for full docstring)."""
        try:
            with model_storage.read_from(resource) as model_dir:
                dataset = rasa.shared.utils.io.read_json_file(
                    model_dir / "crf_dataset.json"
                )
                crf_order = rasa.shared.utils.io.read_json_file(
                    model_dir / "crf_order.json"
                )

                dataset = [
                    [CRFToken.create_from_dict(token_data) for token_data in sub_list]
                    for sub_list in dataset
                ]

                entity_taggers = cls.train_model(dataset, config, crf_order)

                entity_extractor = cls(config, model_storage, resource, entity_taggers)
                entity_extractor.crf_order = crf_order
                return entity_extractor
        except ValueError:
            logger.warning(
                f"Failed to load {cls.__name__} from model storage. Resource "
                f"'{resource.name}' doesn't exist."
            )
            return cls(config, model_storage, resource)

    def persist(self, dataset: List[List[CRFToken]]) -> None:
        """Persist this model into the passed directory."""
        with self._model_storage.write_to(self._resource) as model_dir:
            data_to_store = [
                [token.to_dict() for token in sub_list] for sub_list in dataset
            ]

            rasa.shared.utils.io.dump_obj_as_json_to_file(
                model_dir / "crf_dataset.json", data_to_store
            )
            rasa.shared.utils.io.dump_obj_as_json_to_file(
                model_dir / "crf_order.json", self.crf_order
            )

    @classmethod
    def _crf_tokens_to_features(
        cls,
        crf_tokens: List[CRFToken],
        config: Dict[str, Any],
        include_tag_features: bool = False,
    ) -> List[Dict[Text, Any]]:
        """Convert the list of tokens into discrete features."""
        configured_features = config[CONFIG_FEATURES]
        sentence_features = []

        for token_idx in range(len(crf_tokens)):
            # the features for the current token include features of the token
            # before and after the current features (if defined in the config)
            # token before (-1), current token (0), token after (+1)
            window_size = len(configured_features)
            half_window_size = window_size // 2
            window_range = range(-half_window_size, half_window_size + 1)

            token_features = cls._create_features_for_token(
                crf_tokens,
                token_idx,
                half_window_size,
                window_range,
                include_tag_features,
                config,
            )

            sentence_features.append(token_features)

        return sentence_features

    @classmethod
    def _create_features_for_token(
        cls,
        crf_tokens: List[CRFToken],
        token_idx: int,
        half_window_size: int,
        window_range: range,
        include_tag_features: bool,
        config: Dict[str, Any],
    ) -> Dict[Text, Any]:
        """Convert a token into discrete features including words before and after."""
        configured_features = config[CONFIG_FEATURES]
        prefixes = [str(i) for i in window_range]

        token_features = {}

        # iterate over the tokens in the window range (-1, 0, +1) to collect the
        # features for the token at token_idx
        for pointer_position in window_range:
            current_token_idx = token_idx + pointer_position

            if current_token_idx >= len(crf_tokens):
                # token is at the end of the sentence
                token_features["EOS"] = True
            elif current_token_idx < 0:
                # token is at the beginning of the sentence
                token_features["BOS"] = True
            else:
                token = crf_tokens[current_token_idx]

                # get the features to extract for the token we are currently looking at
                current_feature_idx = pointer_position + half_window_size
                features = configured_features[current_feature_idx]

                prefix = prefixes[current_feature_idx]

                # we add the 'entity' feature to include the entity type as features
                # for the role and group CRFs
                # (do not modify features, otherwise we will end up adding 'entity'
                # over and over again, making training very slow)
                additional_features = []
                if include_tag_features:
                    additional_features.append(CRFEntityExtractorOptions.ENTITY)

                for feature in features + additional_features:
                    if feature == CRFEntityExtractorOptions.PATTERN:
                        # add all regexes extracted from the 'RegexFeaturizer' as a
                        # feature: 'pattern_name' is the name of the pattern the user
                        # set in the training data, 'matched' is either 'True' or
                        # 'False' depending on whether the token actually matches the
                        # pattern or not
                        regex_patterns = cls.function_dict[feature](token)
                        for pattern_name, matched in regex_patterns.items():
                            token_features[f"{prefix}:{feature}:{pattern_name}"] = (
                                matched
                            )
                    else:
                        value = cls.function_dict[feature](token)
                        token_features[f"{prefix}:{feature}"] = value

        return token_features

    @staticmethod
    def _crf_tokens_to_tags(crf_tokens: List[CRFToken], tag_name: Text) -> List[Text]:
        """Return the list of tags for the given tag name."""
        if tag_name == ENTITY_ATTRIBUTE_ROLE:
            return [crf_token.entity_role_tag for crf_token in crf_tokens]
        if tag_name == ENTITY_ATTRIBUTE_GROUP:
            return [crf_token.entity_group_tag for crf_token in crf_tokens]

        return [crf_token.entity_tag for crf_token in crf_tokens]

    @staticmethod
    def _pattern_of_token(message: Message, idx: int) -> Dict[Text, bool]:
        """Get the patterns of the token at the given index extracted by the
        'RegexFeaturizer'.

        The 'RegexFeaturizer' adds all patterns listed in the training data to the
        token. The pattern name is mapped to either 'True' (pattern applies to token) or
        'False' (pattern does not apply to token).

        Args:
            message: The message.
            idx: The token index.

        Returns:
            The pattern dict.
        """
        if message.get(TOKENS_NAMES[TEXT]) is not None:
            return message.get(TOKENS_NAMES[TEXT])[idx].get(
                CRFEntityExtractorOptions.PATTERN, {}
            )
        return {}

    def _get_dense_features(self, message: Message) -> Optional[np.ndarray]:
        """Convert dense features to python-crfsuite feature format."""
        features, _ = message.get_dense_features(
            TEXT, self.component_config["featurizers"]
        )

        if features is None:
            return None

        tokens = message.get(TOKENS_NAMES[TEXT])
        if len(tokens) != len(features.features):
            rasa.shared.utils.io.raise_warning(
                f"Number of dense features ({len(features.features)}) for attribute "
                f"'{TEXT}' does not match number of tokens ({len(tokens)}).",
                docs=DOCS_URL_COMPONENTS + "#crfentityextractor",
            )
            return None

        return features.features

    @staticmethod
    def _convert_dense_features_for_crfsuite(
        crf_token: CRFToken,
    ) -> Dict[Text, Dict[Text, float]]:
        """Converts dense features of CRFTokens to dicts for the crfsuite."""
        feature_dict = {
            str(index): token_features
            for index, token_features in enumerate(crf_token.dense_features)
        }
        converted = {"text_dense_features": feature_dict}
        return converted

    def _convert_to_crf_tokens(self, message: Message) -> List[CRFToken]:
        """Take a message and convert it to crfsuite format."""
        crf_format = []
        tokens = message.get(TOKENS_NAMES[TEXT])

        text_dense_features = self._get_dense_features(message)
        tags = self._get_tags(message)

        for i, token in enumerate(tokens):
            pattern = self._pattern_of_token(message, i)
            entity = self.get_tag_for(tags, ENTITY_ATTRIBUTE_TYPE, i)
            group = self.get_tag_for(tags, ENTITY_ATTRIBUTE_GROUP, i)
            role = self.get_tag_for(tags, ENTITY_ATTRIBUTE_ROLE, i)
            pos_tag = token.get(POS_TAG_KEY)
            dense_features = (
                text_dense_features[i] if text_dense_features is not None else []
            )

            crf_format.append(
                CRFToken(
                    text=token.text,
                    pos_tag=pos_tag,
                    entity_tag=entity,
                    entity_group_tag=group,
                    entity_role_tag=role,
                    pattern=pattern,
                    dense_features=dense_features,
                )
            )

        return crf_format

    def _get_tags(self, message: Message) -> Dict[Text, List[Text]]:
        """Get assigned entity tags of message."""
        tokens = message.get(TOKENS_NAMES[TEXT])
        tags = {}

        for tag_name in self.crf_order:
            if self.component_config[BILOU_FLAG]:
                bilou_key = bilou_utils.get_bilou_key_for_tag(tag_name)
                if message.get(bilou_key):
                    _tags = message.get(bilou_key)
                else:
                    _tags = [NO_ENTITY_TAG for _ in tokens]
            else:
                _tags = [
                    determine_token_labels(
                        token, message.get(ENTITIES), attribute_key=tag_name
                    )
                    for token in tokens
                ]
            tags[tag_name] = _tags

        return tags

    @classmethod
    def train_model(
        cls,
        df_train: List[List[CRFToken]],
        config: Dict[str, Any],
        crf_order: List[str],
    ) -> OrderedDict[str, CRF]:
        """Train the crf tagger based on the training data."""
        import sklearn_crfsuite

        entity_taggers = OrderedDict()

        for tag_name in crf_order:
            logger.debug(f"Training CRF for '{tag_name}'.")

            # add entity tag features for second level CRFs
            include_tag_features = tag_name != ENTITY_ATTRIBUTE_TYPE
            X_train = (
                cls._crf_tokens_to_features(sentence, config, include_tag_features)
                for sentence in df_train
            )
            y_train = (
                cls._crf_tokens_to_tags(sentence, tag_name) for sentence in df_train
            )

            entity_tagger = sklearn_crfsuite.CRF(
                algorithm="lbfgs",
                # coefficient for L1 penalty
                c1=config["L1_c"],
                # coefficient for L2 penalty
                c2=config["L2_c"],
                # stop earlier
                max_iterations=config["max_iterations"],
                # include transitions that are possible, but not observed
                all_possible_transitions=True,
            )
            entity_tagger.fit(X_train, y_train)

            entity_taggers[tag_name] = entity_tagger

            logger.debug("Training finished.")

        return entity_taggers



import os
import shutil
import subprocess
from enum import Enum
from typing import Dict, Optional

import aiohttp
import structlog
from pydantic import BaseModel, ConfigDict

from rasa.constants import MODEL_ARCHIVE_EXTENSION
from rasa.exceptions import ModelNotFound
from rasa.model_manager import config
from rasa.model_manager.utils import (
    logs_path,
    models_base_path,
    subpath,
    write_encoded_data_to_file,
)
from rasa.model_manager.warm_rasa_process import start_rasa_process

structlogger = structlog.get_logger()


class BotSessionStatus(str, Enum):
    """Enum for the bot status."""

    QUEUED = "queued"
    RUNNING = "running"
    STOPPED = "stopped"


class BotSession(BaseModel):
    """Store information about a running bot."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    deployment_id: str
    status: BotSessionStatus
    process: subprocess.Popen
    url: str
    internal_url: str
    port: int
    log_id: str
    returncode: Optional[int] = None

    def is_alive(self) -> bool:
        """Check if the bot is alive."""
        return self.process.poll() is None

    def is_status_indicating_alive(self) -> bool:
        """Check if the status indicates that the bot is alive."""
        return self.status in {BotSessionStatus.QUEUED, BotSessionStatus.RUNNING}

    def has_died_recently(self) -> bool:
        """Check if the bot has died recently.

        Process will indicate that the bot exited,
        but status is not yet updated.
        """
        return self.is_status_indicating_alive() and not self.is_alive()

    async def completed_startup_recently(self) -> bool:
        """Check if the bot has completed startup recently."""
        return self.status == BotSessionStatus.QUEUED and await is_bot_startup_finished(
            self
        )


def bot_path(deployment_id: str) -> str:
    """Return the path to the bot directory for a given deployment id."""
    return os.path.abspath(
        f"{config.SERVER_BASE_WORKING_DIRECTORY}/bots/{deployment_id}"
    )


async def is_bot_startup_finished(bot: BotSession) -> bool:
    """Send a request to the bot to see if the bot is up and running."""
    health_timeout = aiohttp.ClientTimeout(total=5, sock_connect=2, sock_read=3)
    try:
        async with aiohttp.ClientSession(timeout=health_timeout) as session:
            # can't use /status as by default the bot API is not enabled, only
            # the input channel
            async with session.get(f"{bot.internal_url}/license") as resp:
                return resp.status == 200
    except aiohttp.client_exceptions.ClientConnectorError:
        return False


def set_bot_status_to_stopped(bot: BotSession) -> None:
    """Set a bots state to stopped."""
    structlogger.info(
        "model_runner.bot_stopped",
        deployment_id=bot.deployment_id,
        status=bot.process.returncode,
    )
    bot.status = BotSessionStatus.STOPPED
    bot.returncode = bot.process.returncode


def set_bot_status_to_running(bot: BotSession) -> None:
    """Set a bots state to running."""
    structlogger.info(
        "model_runner.bot_running",
        deployment_id=bot.deployment_id,
    )
    bot.status = BotSessionStatus.RUNNING


def get_open_port() -> int:
    """Get an open port on the system that is not in use yet."""
    # from https://stackoverflow.com/questions/2838244/get-open-tcp-port-in-python/2838309#2838309
    import socket

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


def write_encoded_config_data_to_files(
    encoded_configs: Dict[str, bytes], base_path: str
) -> None:
    """Write the encoded config data to files."""
    for key, value in encoded_configs.items():
        write_encoded_data_to_file(value, subpath(base_path, f"{key}.yml"))


def prepare_bot_directory(
    bot_base_path: str,
    model_name: str,
    encoded_configs: Dict[str, str],
) -> None:
    """Prepare the bot directory for a new bot session."""
    if not os.path.exists(bot_base_path):
        os.makedirs(bot_base_path, exist_ok=True)
    else:
        shutil.rmtree(bot_base_path, ignore_errors=True)

    model_file_name = f"{model_name}.{MODEL_ARCHIVE_EXTENSION}"
    model_path = subpath(models_base_path(), model_file_name)

    if config.SERVER_MODEL_REMOTE_STORAGE and not os.path.exists(model_path):
        fetch_remote_model_to_dir(
            model_file_name,
            models_base_path(),
            config.SERVER_MODEL_REMOTE_STORAGE,
        )

    if not os.path.exists(model_path):
        raise ModelNotFound(f"Model '{model_file_name}' not found in '{model_path}'.")

    os.makedirs(subpath(bot_base_path, "models"), exist_ok=True)
    shutil.copy(
        src=model_path,
        dst=subpath(bot_base_path, "models"),
    )

    write_encoded_config_data_to_files(encoded_configs, bot_base_path)


def fetch_remote_model_to_dir(
    model_name: str, target_path: str, storage_type: str
) -> str:
    """Fetch the model from remote storage.

    Returns the path to the model diretory.
    """
    from rasa.core.persistor import get_persistor

    persistor = get_persistor(storage_type)

    # we now there must be a persistor, because the config is set
    # this is here to please the type checker for the call below
    assert persistor is not None

    try:
        return persistor.retrieve(model_name=model_name, target_path=target_path)
    except FileNotFoundError as e:
        raise ModelNotFound() from e


def fetch_size_of_remote_model(model_name: str, storage_type: str) -> int:
    """Fetch the size of the model from remote storage."""
    from rasa.core.persistor import get_persistor

    persistor = get_persistor(storage_type)

    # we now there must be a persistor, because the config is set
    # this is here to please the type checker for the call below
    assert persistor is not None

    return persistor.size_of_persisted_model(model_name=model_name)


def start_bot_process(
    deployment_id: str, bot_base_path: str, base_url_path: str
) -> BotSession:
    port = get_open_port()

    arguments = [
        "run",
        "--endpoints",
        f"{bot_base_path}/endpoints.yml",
        "--credentials",
        f"{bot_base_path}/credentials.yml",
        "--debug",
        f"--port={port}",
        "--model",
        f"{bot_base_path}/models",
    ]

    structlogger.debug(
        "model_runner.bot.starting_command",
        deployment_id=deployment_id,
        arguments=" ".join(arguments),
    )

    warm_process = start_rasa_process(cwd=bot_base_path, arguments=arguments)

    internal_bot_url = f"http://localhost:{port}"

    structlogger.info(
        "model_runner.bot.starting",
        deployment_id=deployment_id,
        log=logs_path(warm_process.log_id),
        url=internal_bot_url,
        port=port,
        pid=warm_process.process.pid,
    )

    return BotSession(
        deployment_id=deployment_id,
        status=BotSessionStatus.QUEUED,
        process=warm_process.process,
        url=f"{base_url_path}?deployment_id={deployment_id}",
        internal_url=internal_bot_url,
        port=port,
        log_id=warm_process.log_id,
    )


def run_bot(
    deployment_id: str,
    model_name: str,
    base_url_path: str,
    encoded_configs: Dict[str, str],
) -> BotSession:
    """Deploy a bot based on a given training id."""
    with structlog.contextvars.bound_contextvars(model_name=model_name):
        bot_base_path = bot_path(deployment_id)
        prepare_bot_directory(bot_base_path, model_name, encoded_configs)

        return start_bot_process(deployment_id, bot_base_path, base_url_path)


async def update_bot_status(bot: BotSession) -> None:
    """Update the status of a bot based on the process return code."""
    try:
        if bot.has_died_recently():
            set_bot_status_to_stopped(bot)
        elif await bot.completed_startup_recently():
            set_bot_status_to_running(bot)
    except Exception as e:
        structlogger.error("model_runner.update_bot_status.error", error=str(e))


def terminate_bot(bot: BotSession) -> None:
    """Terminate the bot process."""
    if not bot.is_status_indicating_alive():
        # if the bot is not running, we don't need to terminate it
        return

    try:
        bot.process.terminate()
        structlogger.info(
            "model_runner.stop_bot.stopped",
            deployment_id=bot.deployment_id,
            status=bot.process.returncode,
        )
        bot.status = BotSessionStatus.STOPPED
        bot.returncode = bot.process.returncode
    except ProcessLookupError:
        structlogger.debug(
            "model_runner.stop_bot.process_not_found",
            deployment_id=bot.deployment_id,
        )



import json
import logging
from collections import defaultdict
from typing import DefaultDict, Dict, Generator, List, NamedTuple, Optional, Text, Tuple

from rasa.core.featurizers.tracker_featurizers import MaxHistoryTrackerFeaturizer
from rasa.nlu.tokenizers.tokenizer import Tokenizer
from rasa.shared.core.constants import (
    ACTION_LISTEN_NAME,
    ACTION_UNLIKELY_INTENT_NAME,
    PREVIOUS_ACTION,
    USER,
)
from rasa.shared.core.domain import Domain, State
from rasa.shared.core.events import ActionExecuted, Event
from rasa.shared.core.generator import TrackerWithCachedStates
from rasa.shared.nlu.constants import TEXT
from rasa.shared.nlu.training_data.message import Message

logger = logging.getLogger(__name__)


class StoryConflict:
    """Represents a conflict between two or more stories.

    Here, a conflict means that different actions are supposed to follow from
    the same dialogue state, which most policies cannot learn.
    """

    def __init__(self, sliced_states: List[State]) -> None:
        """Creates a `StoryConflict` from a given state.

        Args:
            sliced_states: The (sliced) dialogue state at which the conflict occurs.
        """
        self._sliced_states = sliced_states
        # A list of actions that all follow from the same state.
        self._conflicting_actions: DefaultDict[Text, List[Text]] = defaultdict(
            list
        )  # {"action": ["story_1", ...], ...}

    def __hash__(self) -> int:
        return hash(str(list(self._sliced_states)))

    def add_conflicting_action(self, action: Text, story_name: Text) -> None:
        """Adds another action that follows from the same state.

        Args:
            action: Name of the action.
            story_name: Name of the story where this action is chosen.
        """
        self._conflicting_actions[action] += [story_name]

    @property
    def conflicting_actions(self) -> List[Text]:
        """List of conflicting actions.

        Returns:
            List of conflicting actions.

        """
        return list(self._conflicting_actions.keys())

    @property
    def conflict_has_prior_events(self) -> bool:
        """Checks if prior events exist.

        Returns:
            `True` if anything has happened before this conflict, otherwise `False`.
        """
        return _get_previous_event(self._sliced_states[-1])[0] is not None

    def __str__(self) -> Text:
        # Describe where the conflict occurs in the stories
        last_event_type, last_event_name = _get_previous_event(self._sliced_states[-1])
        if last_event_type:
            conflict_message = (
                f"Story structure conflict after {last_event_type} "
                f"'{last_event_name}':\n"
            )
        else:
            conflict_message = "Story structure conflict at the beginning of stories:\n"

        # List which stories are in conflict with one another
        for action, stories in self._conflicting_actions.items():
            conflict_message += (
                f"  {self._summarize_conflicting_actions(action, stories)}"
            )

        return conflict_message

    @staticmethod
    def _summarize_conflicting_actions(action: Text, stories: List[Text]) -> Text:
        """Gives a summarized textual description of where one action occurs.

        Args:
            action: The name of the action.
            stories: The stories in which the action occurs.

        Returns:
            A textural summary.
        """
        if len(stories) > 3:
            # Four or more stories are present
            conflict_description = (
                f"'{stories[0]}', '{stories[1]}', and {len(stories) - 2} other trackers"
            )
        elif len(stories) == 3:
            conflict_description = f"'{stories[0]}', '{stories[1]}', and '{stories[2]}'"
        elif len(stories) == 2:
            conflict_description = f"'{stories[0]}' and '{stories[1]}'"
        elif len(stories) == 1:
            conflict_description = f"'{stories[0]}'"
        else:
            raise ValueError(
                "An internal error occurred while trying to summarise a conflict "
                "without stories. Please file a bug report at "
                "https://github.com/RasaHQ/rasa."
            )

        return f"{action} predicted in {conflict_description}\n"


class TrackerEventStateTuple(NamedTuple):
    """Holds a tracker, an event, and sliced states associated with those."""

    tracker: TrackerWithCachedStates
    event: Event
    sliced_states: List[State]

    @property
    def sliced_states_hash(self) -> int:
        """Returns the hash of the sliced states."""
        return hash(json.dumps(self.sliced_states, sort_keys=True))


def find_story_conflicts(
    trackers: List[TrackerWithCachedStates],
    domain: Domain,
    max_history: Optional[int] = None,
) -> List[StoryConflict]:
    """Generates `StoryConflict` objects, describing conflicts in the given trackers.

    Args:
        trackers: Trackers in which to search for conflicts.
        domain: The domain.
        max_history: The maximum history length to be taken into account.

    Returns:
        StoryConflict objects.
    """
    if max_history:
        logger.info(
            f"Considering the preceding {max_history} turns for conflict analysis."
        )
    else:
        logger.info("Considering all preceding turns for conflict analysis.")

    # We do this in two steps, to reduce memory consumption:

    # Create a 'state -> list of actions' dict, where the state is
    # represented by its hash
    conflicting_state_action_mapping = _find_conflicting_states(
        trackers, domain, max_history
    )

    # Iterate once more over all states and note the (unhashed) state,
    # for which a conflict occurs
    conflicts = _build_conflicts_from_states(
        trackers, domain, max_history, conflicting_state_action_mapping
    )

    return conflicts


def _find_conflicting_states(
    trackers: List[TrackerWithCachedStates],
    domain: Domain,
    max_history: Optional[int],
    tokenizer: Optional[Tokenizer] = None,
) -> Dict[int, List[int]]:
    """Identifies all states from which different actions follow.

    Args:
        trackers: Trackers that contain the states.
        domain: The domain object.
        max_history: Number of turns to take into account for the state descriptions.
        tokenizer: A tokenizer to tokenize the user messages.

    Returns:
        A dictionary mapping state-hashes to a list of actions that follow from each
        state.
    """
    # Create a 'state -> list of actions' dict, where the state is
    # represented by its hash
    state_action_mapping: DefaultDict[int, List[int]] = defaultdict(list)

    for element in _sliced_states_iterator(trackers, domain, max_history, tokenizer):
        hashed_state = element.sliced_states_hash
        current_hash = hash(element.event)

        if current_hash not in state_action_mapping[
            hashed_state
        ] or _unlearnable_action(element.event):
            state_action_mapping[hashed_state] += [current_hash]

    # Keep only conflicting `state_action_mapping`s
    # or those mappings that contain `action_unlikely_intent`
    action_unlikely_intent_hash = hash(
        ActionExecuted(action_name=ACTION_UNLIKELY_INTENT_NAME)
    )
    return {
        state_hash: actions
        for (state_hash, actions) in state_action_mapping.items()
        if len(actions) > 1 or action_unlikely_intent_hash in actions
    }


def _unlearnable_action(event: Event) -> bool:
    """Identifies if the action cannot be learned by policies that use story data.

    Args:
        event: An event to be checked.

    Returns:
        `True` if the event can be learned, `False` otherwise.
    """
    return (
        isinstance(event, ActionExecuted)
        and event.action_name == ACTION_UNLIKELY_INTENT_NAME
    )


def _build_conflicts_from_states(
    trackers: List[TrackerWithCachedStates],
    domain: Domain,
    max_history: Optional[int],
    conflicting_state_action_mapping: Dict[int, List[int]],
    tokenizer: Optional[Tokenizer] = None,
) -> List["StoryConflict"]:
    """Builds a list of `StoryConflict` objects for each given conflict.

    Args:
        trackers: Trackers that contain the states.
        domain: The domain object.
        max_history: Number of turns to take into account for the state descriptions.
        conflicting_state_action_mapping: A dictionary mapping state-hashes to a list
            of actions that follow from each state.
        tokenizer: A tokenizer to tokenize the user messages.

    Returns:
        A list of `StoryConflict` objects that describe inconsistencies in the story
        structure. These objects also contain the history that leads up to the conflict.
    """
    # Iterate once more over all states and note the (unhashed) state,
    # for which a conflict occurs
    conflicts = {}
    for element in _sliced_states_iterator(trackers, domain, max_history, tokenizer):
        hashed_state = element.sliced_states_hash

        if hashed_state in conflicting_state_action_mapping:
            if hashed_state not in conflicts:
                conflicts[hashed_state] = StoryConflict(element.sliced_states)

            conflicts[hashed_state].add_conflicting_action(
                action=str(element.event), story_name=element.tracker.sender_id
            )

    # Return list of conflicts that arise from unpredictable actions
    # (actions that start the conversation)
    return [
        conflict
        for (hashed_state, conflict) in conflicts.items()
        if conflict.conflict_has_prior_events
    ]


def _sliced_states_iterator(
    trackers: List[TrackerWithCachedStates],
    domain: Domain,
    max_history: Optional[int],
    tokenizer: Optional[Tokenizer],
) -> Generator[TrackerEventStateTuple, None, None]:
    """Creates an iterator over sliced states.

    Iterate over all given trackers and all sliced states within each tracker,
    where the slicing is based on `max_history`.

    Args:
        trackers: List of trackers.
        domain: Domain (used for tracker.past_states).
        max_history: Assumed `max_history` value for slicing.
        tokenizer: A tokenizer to tokenize the user messages.

    Yields:
        A (tracker, event, sliced_states) triplet.
    """
    for tracker in trackers:
        states = tracker.past_states(domain)

        idx = 0
        for event in tracker.events:
            if isinstance(event, ActionExecuted):
                sliced_states = MaxHistoryTrackerFeaturizer.slice_state_history(
                    states[: idx + 1], max_history
                )
                if tokenizer:
                    _apply_tokenizer_to_states(tokenizer, sliced_states)
                # TODO: deal with oov (different tokens can lead to identical features
                # if some of those tokens are out of vocabulary for all featurizers)
                yield TrackerEventStateTuple(tracker, event, sliced_states)
                idx += 1


def _apply_tokenizer_to_states(tokenizer: Tokenizer, states: List[State]) -> None:
    """Split each user text into tokens and concatenate them again.

    Args:
        tokenizer: A tokenizer to tokenize the user messages.
        states: The states to be tokenized.
    """
    for state in states:
        if USER in state and TEXT in state[USER]:
            state[USER][TEXT] = " ".join(
                token.text
                for token in tokenizer.tokenize(
                    Message({TEXT: state[USER][TEXT]}), TEXT
                )
            )


def _get_previous_event(
    state: Optional[State],
) -> Tuple[Optional[Text], Optional[Text]]:
    """Returns previous event type and name.

    Returns the type and name of the event (action or intent) previous to the
    given state (excluding action_listen).

    Args:
        state: Element of sliced states.

    Returns:
        Tuple of (type, name) strings of the prior event.
    """
    previous_event_type = None
    previous_event_name = None

    # A typical state might be
    # `{'user': {'intent': 'greet'}, 'prev_action': {'action_name': 'action_listen'}}`.
    if not state:
        previous_event_type = None
        previous_event_name = None
    elif (
        PREVIOUS_ACTION in state.keys()
        and "action_name" in state[PREVIOUS_ACTION]
        and state[PREVIOUS_ACTION]["action_name"] != ACTION_LISTEN_NAME
    ):
        previous_event_type = "action"
        previous_event_name = state[PREVIOUS_ACTION]["action_name"]
    elif PREVIOUS_ACTION in state.keys() and "action_text" in state[PREVIOUS_ACTION]:
        previous_event_type = "bot utterance"
        previous_event_name = state[PREVIOUS_ACTION]["action_text"]
    elif USER in state.keys():
        if "intent" in state[USER]:
            previous_event_type = "intent"
            previous_event_name = state[USER]["intent"]
        elif "text" in state[USER]:
            previous_event_type = "user utterance"
            previous_event_name = state[USER]["text"]

    if not isinstance(previous_event_name, (str, type(None))):
        # While the Substate type doesn't restrict the value of `action_text` /
        # `intent`, etc. to be a string, it always should be
        raise TypeError(
            f"The value '{previous_event_name}' in the substate should be a string or "
            f"None, not {type(previous_event_name)}. Did you modify Rasa source code?"
        )

    return previous_event_type, previous_event_name



import asyncio
import json
from typing import Any, Awaitable, Callable, Dict, Optional, Text

import structlog
from sanic import (  # type: ignore[attr-defined]
    Blueprint,
    HTTPResponse,
    Request,
    Websocket,
    response,
)

from rasa.core.channels import UserMessage
from rasa.core.channels.voice_ready.utils import CallParameters
from rasa.core.channels.voice_stream.audio_bytes import RasaAudioBytes
from rasa.core.channels.voice_stream.call_state import (
    call_state,
)
from rasa.core.channels.voice_stream.tts.tts_engine import TTSEngine
from rasa.core.channels.voice_stream.voice_channel import (
    ContinueConversationAction,
    EndConversationAction,
    NewAudioAction,
    VoiceChannelAction,
    VoiceInputChannel,
    VoiceOutputChannel,
)

# Not mentioned in the documentation but observed in Geneys's example
# https://github.com/GenesysCloudBlueprints/audioconnector-server-reference-implementation
MAXIMUM_BINARY_MESSAGE_SIZE = 64000  # 64KB
logger = structlog.get_logger(__name__)


def map_call_params(data: Dict[Text, Any]) -> CallParameters:
    """Map the twilio stream parameters to the CallParameters dataclass."""
    parameters = data["parameters"]
    participant = parameters["participant"]
    # sent as {"ani": "tel:+491604697810"}
    ani = participant.get("ani", "")
    user_phone = ani.split(":")[-1] if ani else ""

    return CallParameters(
        call_id=parameters.get("conversationId", ""),
        user_phone=user_phone,
        bot_phone=participant.get("dnis", ""),
    )


class GenesysOutputChannel(VoiceOutputChannel):
    @classmethod
    def name(cls) -> str:
        return "genesys"

    async def send_audio_bytes(
        self, recipient_id: str, audio_bytes: RasaAudioBytes
    ) -> None:
        """
        Send audio bytes to the recipient with buffering.

        Genesys throws a rate limit error with too many audio messages.
        To avoid this, we buffer the audio messages and send them in chunks.

        - global.inbound.binary.average.rate.per.second: 5
        The allowed average rate per second of inbound binary data

        - global.inbound.binary.max: 25
        The maximum number of inbound binary data messages
        that can be sent instantaneously

        https://developer.genesys.cloud/organization/organization/limits#audiohook
        """
        call_state.audio_buffer.extend(audio_bytes)

        # If we receive a non-standard chunk size, assume it's the end of a sequence
        # or buffer is more than 32KB (this is half of genesys's max audio message size)
        if len(audio_bytes) != 1024 or len(call_state.audio_buffer) >= (
            MAXIMUM_BINARY_MESSAGE_SIZE / 2
        ):
            # TODO: we should send the buffer when we receive a synthesis complete event
            # from TTS. This will ensure that the last audio chunk is always sent.
            await self._send_audio_buffer(self.voice_websocket)

    async def _send_audio_buffer(self, ws: Websocket) -> None:
        """Send the audio buffer to the recipient if it's not empty."""
        if call_state.audio_buffer:
            buffer_bytes = bytes(call_state.audio_buffer)
            await self._send_bytes_to_ws(ws, buffer_bytes)
            call_state.audio_buffer.clear()

    async def _send_bytes_to_ws(self, ws: Websocket, data: bytes) -> None:
        """Send audio bytes to the recipient as a binary websocket message."""
        if len(data) <= MAXIMUM_BINARY_MESSAGE_SIZE:
            await self.voice_websocket.send(data)
        else:
            # split the audio into chunks
            current_position = 0
            while current_position < len(data):
                end_position = min(
                    current_position + MAXIMUM_BINARY_MESSAGE_SIZE, len(data)
                )
                await self.voice_websocket.send(data[current_position:end_position])
                current_position = end_position

    async def send_marker_message(self, recipient_id: str) -> None:
        """
        Send a message that marks positions in the audio stream.
        Genesys does not support this feature, so we do nothing here.
        """
        pass


class GenesysInputChannel(VoiceInputChannel):
    @classmethod
    def name(cls) -> str:
        return "genesys"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def _get_next_sequence(self) -> int:
        """
        Get the next message sequence number
        Rasa == Server
        Genesys == Client

        Genesys requires the server and client each maintain a
        monotonically increasing message sequence number.
        """
        cs = call_state
        cs.server_sequence_number += 1  # type: ignore[attr-defined]
        return cs.server_sequence_number

    def _get_last_client_sequence(self) -> int:
        """Get the last client(Genesys) sequence number."""
        return call_state.client_sequence_number

    def _update_client_sequence(self, seq: int) -> None:
        """Update the client(Genesys) sequence number."""
        if seq - call_state.client_sequence_number != 1:
            logger.warning(
                "genesys.update_client_sequence.sequence_gap",
                received_seq=seq,
                last_seq=call_state.client_sequence_number,
            )
        call_state.client_sequence_number = seq  # type: ignore[attr-defined]

    def channel_bytes_to_rasa_audio_bytes(self, input_bytes: bytes) -> RasaAudioBytes:
        return RasaAudioBytes(input_bytes)

    async def collect_call_parameters(
        self, channel_websocket: Websocket
    ) -> Optional[CallParameters]:
        """Call Parameters are collected during the open event."""
        async for message in channel_websocket:
            data = json.loads(message)
            self._update_client_sequence(data["seq"])
            if data.get("type") == "open":
                call_params = await self.handle_open(channel_websocket, data)
                return call_params
            else:
                logger.error("genesys.receive.unexpected_initial_message", message=data)

        return None

    def map_input_message(
        self,
        message: Any,
        ws: Websocket,
    ) -> VoiceChannelAction:
        # if message is binary, it's audio
        if isinstance(message, bytes):
            return NewAudioAction(self.channel_bytes_to_rasa_audio_bytes(message))
        else:
            # process text message
            data = json.loads(message)
            self._update_client_sequence(data["seq"])
            msg_type = data.get("type")
            if msg_type == "close":
                logger.info("genesys.handle_close", message=data)
                self.handle_close(ws, data)
                return EndConversationAction()
            elif msg_type == "ping":
                logger.info("genesys.handle_ping", message=data)
                self.handle_ping(ws, data)
            elif msg_type == "playback_started":
                logger.debug("genesys.handle_playback_started", message=data)
                call_state.is_bot_speaking = True  # type: ignore[attr-defined]
            elif msg_type == "playback_completed":
                logger.debug("genesys.handle_playback_completed", message=data)
                call_state.is_bot_speaking = False  # type: ignore[attr-defined]
                if call_state.should_hangup:
                    logger.info("genesys.hangup")
                    self.disconnect(ws, data)
                    # the conversation should continue until
                    # we receive a close message from Genesys
            elif msg_type == "dtmf":
                logger.info("genesys.handle_dtmf", message=data)
            elif msg_type == "error":
                logger.warning("genesys.handle_error", message=data)
            else:
                logger.warning("genesys.map_input_message.unknown_type", message=data)

        return ContinueConversationAction()

    def create_output_channel(
        self, voice_websocket: Websocket, tts_engine: TTSEngine
    ) -> VoiceOutputChannel:
        return GenesysOutputChannel(
            voice_websocket,
            tts_engine,
            self.tts_cache,
        )

    async def handle_open(self, ws: Websocket, message: dict) -> CallParameters:
        """Handle initial open transaction from Genesys."""
        call_parameters = map_call_params(message)
        params = message["parameters"]
        media_options = params.get("media", [])

        # Send opened response
        if media_options:
            logger.info("genesys.handle_open", media_parameter=media_options[0])
            response = {
                "version": "2",
                "type": "opened",
                "seq": self._get_next_sequence(),
                "clientseq": self._get_last_client_sequence(),
                "id": message.get("id"),
                "parameters": {"startPaused": False, "media": [media_options[0]]},
            }
            logger.debug("genesys.handle_open.opened", response=response)
            await ws.send(json.dumps(response))
        else:
            logger.warning(
                "genesys.handle_open.no_media_formats", client_message=message
            )
        return call_parameters

    def handle_ping(self, ws: Websocket, message: dict) -> None:
        """Handle ping message from Genesys."""
        response = {
            "version": "2",
            "type": "pong",
            "seq": self._get_next_sequence(),
            "clientseq": message.get("seq"),
            "id": message.get("id"),
            "parameters": {},
        }
        logger.debug("genesys.handle_ping.pong", response=response)
        _schedule_ws_task(ws.send(json.dumps(response)))

    def handle_close(self, ws: Websocket, message: dict) -> None:
        """Handle close message from Genesys."""
        response = {
            "version": "2",
            "type": "closed",
            "seq": self._get_next_sequence(),
            "clientseq": self._get_last_client_sequence(),
            "id": message.get("id"),
            "parameters": message.get("parameters", {}),
        }
        logger.debug("genesys.handle_close.closed", response=response)

        _schedule_ws_task(ws.send(json.dumps(response)))

    def disconnect(self, ws: Websocket, data: dict) -> None:
        """
        Send disconnect message to Genesys.

        https://developer.genesys.cloud/devapps/audiohook/protocol-reference#disconnect
        It should be used to hangup the call.
        Genesys will respond with a "close" message to us
        that is handled by the handle_close method.
        """
        message = {
            "version": "2",
            "type": "disconnect",
            "seq": self._get_next_sequence(),
            "clientseq": self._get_last_client_sequence(),
            "id": data.get("id"),
            "parameters": {
                "reason": "completed",
                # arbitrary values can be sent here
            },
        }
        logger.debug("genesys.disconnect", message=message)
        _schedule_ws_task(ws.send(json.dumps(message)))

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        """Defines a Sanic blueprint for the voice input channel."""
        blueprint = Blueprint("genesys", __name__)

        @blueprint.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @blueprint.websocket("/websocket")  # type: ignore[misc]
        async def receive(request: Request, ws: Websocket) -> None:
            logger.debug(
                "genesys.receive",
                audiohook_session_id=request.headers.get("audiohook-session-id"),
            )
            # validate required headers
            required_headers = [
                "audiohook-organization-id",
                "audiohook-correlation-id",
                "audiohook-session-id",
                "x-api-key",
            ]

            for header in required_headers:
                if header not in request.headers:
                    await ws.close(1008, f"Missing required header: {header}")
                    return

            # TODO: validate API key header
            # process audio streaming
            logger.info("genesys.receive", message="Starting audio streaming")
            await self.run_audio_streaming(on_new_message, ws)

        return blueprint


def _schedule_ws_task(coro: Awaitable[Any]) -> None:
    """Helper function to schedule a coroutine in the event loop.

    Args:
        coro: The coroutine to schedule
    """
    loop = asyncio.get_running_loop()
    loop.call_soon_threadsafe(lambda: loop.create_task(coro))



from __future__ import annotations

import os
from typing import Any, Dict, List

import structlog
from litellm import Router

from rasa.shared.constants import (
    API_KEY,
    LITELLM_PARAMS_KEY,
    MODEL_CONFIG_KEY,
    MODEL_GROUP_ID_CONFIG_KEY,
    MODEL_LIST_KEY,
    ROUTER_CONFIG_KEY,
    SELF_HOSTED_VLLM_API_KEY_ENV_VAR,
    SELF_HOSTED_VLLM_PREFIX,
    USE_CHAT_COMPLETIONS_ENDPOINT_CONFIG_KEY,
)
from rasa.shared.exceptions import ProviderClientValidationError
from rasa.shared.providers._configs.litellm_router_client_config import (
    LiteLLMRouterClientConfig,
)
from rasa.shared.utils.io import resolve_environment_variables

structlogger = structlog.get_logger()


class _BaseLiteLLMRouterClient:
    """An abstract base class for LiteLLM Router clients.

    This class defines the interface and common functionality for all the router clients
    based on LiteLLM.

    The class is made private to prevent it from being part of the public-facing
    interface, as it serves as an internal base class for specific implementations of
    router clients that are based on LiteLLM router implementation.

    Parameters:
        model_group_id (str): The model group ID.
        model_configurations (List[Dict[str, Any]]): The list of model configurations.
        router_settings (Dict[str, Any]): The router settings.
        kwargs (Optional[Dict[str, Any]]): Additional configuration parameters.

    Raises:
        ProviderClientValidationError: If validation of the client setup fails.
    """

    def __init__(
        self,
        model_group_id: str,
        model_configurations: List[Dict[str, Any]],
        router_settings: Dict[str, Any],
        use_chat_completions_endpoint: bool = True,
        **kwargs: Any,
    ):
        self._model_group_id = model_group_id
        self._model_configurations = model_configurations
        self._router_settings = router_settings
        self._use_chat_completions_endpoint = use_chat_completions_endpoint
        self._extra_parameters = kwargs or {}
        self.additional_client_setup()
        try:
            resolved_model_configurations = (
                self._resolve_env_vars_in_model_configurations()
            )
            self._router_client = Router(
                model_list=resolved_model_configurations, **router_settings
            )
        except Exception as e:
            event_info = "Cannot instantiate a router client."
            structlogger.error(
                "_base_litellm_router_client.init.error",
                event_info=event_info,
                model_group_id=model_group_id,
                models=model_configurations,
                router=router_settings,
                original_error=e,
            )
            raise ProviderClientValidationError(f"{event_info} Original error: {e}")

    def additional_client_setup(self) -> None:
        """Additional setup for the LiteLLM Router client."""
        # If the model configuration is self-hosted VLLM, set a dummy API key if not
        # provided. A bug in the LiteLLM library requires an API key to be set even if
        # it is not required.
        for model_configuration in self.model_configurations:
            if (
                f"{SELF_HOSTED_VLLM_PREFIX}/"
                in model_configuration[LITELLM_PARAMS_KEY][MODEL_CONFIG_KEY]
                and API_KEY not in model_configuration[LITELLM_PARAMS_KEY]
                and not os.getenv(SELF_HOSTED_VLLM_API_KEY_ENV_VAR)
            ):
                os.environ[SELF_HOSTED_VLLM_API_KEY_ENV_VAR] = "dummy api key"
                return

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> _BaseLiteLLMRouterClient:
        """Instantiates a LiteLLM Router Embedding client from a configuration dict.

        Args:
            config: (Dict[str, Any]) The configuration dictionary.

        Returns:
            LiteLLMRouterLLMClient: The instantiated LiteLLM Router LLM client.

        Raises:
            ValueError: If the configuration is invalid.
        """
        try:
            client_config = LiteLLMRouterClientConfig.from_dict(config)
        except (KeyError, ValueError) as e:
            message = "Cannot instantiate a client from the passed configuration."
            structlogger.error(
                "litellm_router_llm_client.from_config.error",
                message=message,
                config=config,
                original_error=e,
            )
            raise

        return cls(
            model_group_id=client_config.model_group_id,
            model_configurations=client_config.litellm_model_list,
            router_settings=client_config.litellm_router_settings,
            use_chat_completions_endpoint=client_config.use_chat_completions_endpoint,
            **client_config.extra_parameters,
        )

    @property
    def model_group_id(self) -> str:
        """Returns the model group ID for the LiteLLM Router client."""
        return self._model_group_id

    @property
    def model_configurations(self) -> List[Dict[str, Any]]:
        """Returns the model configurations for the LiteLLM Router client."""
        return self._model_configurations

    @property
    def router_settings(self) -> Dict[str, Any]:
        """Returns the router settings for the LiteLLM Router client."""
        return self._router_settings

    @property
    def router_client(self) -> Router:
        """Returns the instantiated LiteLLM Router client."""
        return self._router_client

    @property
    def use_chat_completions_endpoint(self) -> bool:
        """Returns whether to use the chat completions endpoint."""
        return self._use_chat_completions_endpoint

    @property
    def _litellm_extra_parameters(self) -> Dict[str, Any]:
        """
        Returns the extra parameters for the LiteLLM Router client.

        Returns:
            Dictionary containing the model parameters.
        """
        return self._extra_parameters

    @property
    def config(self) -> Dict:
        """Returns the configuration for the LiteLLM Router client in LiteLLM format."""
        return {
            MODEL_GROUP_ID_CONFIG_KEY: self.model_group_id,
            MODEL_LIST_KEY: self.model_configurations,
            ROUTER_CONFIG_KEY: self.router_settings,
            USE_CHAT_COMPLETIONS_ENDPOINT_CONFIG_KEY: (
                self.use_chat_completions_endpoint
            ),
            **self._litellm_extra_parameters,
        }

    def _resolve_env_vars_in_model_configurations(self) -> List:
        model_configuration_with_resolved_keys = []
        for model_configuration in self.model_configurations:
            resolved_model_configuration = resolve_environment_variables(
                model_configuration
            )
            model_configuration_with_resolved_keys.append(resolved_model_configuration)
        return model_configuration_with_resolved_keys



from typing import Iterable, List, Optional, Text

from rasa.shared.core.domain import Domain
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.training_data.structures import StoryGraph
from rasa.shared.nlu.training_data.training_data import TrainingData


def training_data_from_paths(paths: Iterable[Text], language: Text) -> TrainingData:
    from rasa.shared.nlu.training_data import loading

    training_data_sets = [loading.load_data(nlu_file, language) for nlu_file in paths]
    return TrainingData().merge(*training_data_sets)


def story_graph_from_paths(
    files: List[Text], domain: Domain, exclusion_percentage: Optional[int] = None
) -> StoryGraph:
    """Returns the `StoryGraph` from paths."""
    from rasa.shared.core.training_data import loading

    story_steps = loading.load_data_from_files(files, domain, exclusion_percentage)
    return StoryGraph(story_steps)


def flows_from_paths(files: List[Text]) -> FlowsList:
    """Returns the flows from paths."""
    from rasa.shared.core.flows.yaml_flows_io import YAMLFlowsReader

    flows = FlowsList(underlying_flows=[])
    for file in files:
        flows = flows.merge(
            YAMLFlowsReader.read_from_file(file), ignore_duplicates=False
        )
    flows.validate()
    return flows



from typing import Any, Dict, List, Optional, Text

import structlog

import rasa.shared.utils.io
from rasa.dialogue_understanding.commands import (
    CannotHandleCommand,
    Command,
    ErrorCommand,
    SetSlotCommand,
)
from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxManager,
    CommandSyntaxVersion,
)
from rasa.dialogue_understanding.generator import LLMBasedCommandGenerator
from rasa.dialogue_understanding.generator.command_parser import (
    parse_commands as parse_commands_using_command_parsers,
)
from rasa.dialogue_understanding.generator.constants import (
    COMMAND_PROMPT_FILE_NAME,
    DEFAULT_LLM_CONFIG,
    FLOW_RETRIEVAL_KEY,
    LLM_BASED_COMMAND_GENERATOR_CONFIG_FILE,
    LLM_CONFIG_KEY,
    USER_INPUT_CONFIG_KEY,
)
from rasa.dialogue_understanding.generator.flow_retrieval import FlowRetrieval
from rasa.dialogue_understanding.stack.utils import top_flow_frame
from rasa.dialogue_understanding.utils import (
    add_commands_to_message_parse_data,
    add_prompt_to_message_parse_data,
)
from rasa.engine.graph import ExecutionContext
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.constants import (
    EMBEDDINGS_CONFIG_KEY,
    PROMPT_TEMPLATE_CONFIG_KEY,
    ROUTE_TO_CALM_SLOT,
)
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import ProviderClientAPIException
from rasa.shared.nlu.constants import LLM_COMMANDS, LLM_PROMPT, TEXT
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.providers.llm.llm_response import LLMResponse
from rasa.shared.utils.io import deep_container_fingerprint
from rasa.shared.utils.llm import (
    allowed_values_for_slot,
    get_default_prompt_template_based_on_model,
    get_prompt_template,
    resolve_model_client_config,
    sanitize_message_for_prompt,
    tracker_as_readable_transcript,
)
from rasa.utils.beta import BetaNotEnabledException, ensure_beta_feature_is_enabled
from rasa.utils.log_utils import log_llm

structlogger = structlog.get_logger()


MODEL_PROMPT_MAPPER = {
    "openai/gpt-4o-2024-11-20": "command_prompt_v2_gpt_4o_2024_11_20_template.jinja2",
    "azure/gpt-4o-2024-11-20": "command_prompt_v2_gpt_4o_2024_11_20_template.jinja2",
    "bedrock/anthropic.claude-3-5-sonnet-20240620-v1:0": (
        "command_prompt_v2_claude_3_5_sonnet_20240620_template.jinja2"
    ),
    "anthropic/claude-3-5-sonnet-20240620": (
        "command_prompt_v2_claude_3_5_sonnet_20240620_template.jinja2"
    ),
}


DEFAULT_COMMAND_PROMPT_TEMPLATE_FILE_NAME = "command_prompt_v2_default.jinja2"


@DefaultV1Recipe.register(
    [
        DefaultV1Recipe.ComponentType.COMMAND_GENERATOR,
    ],
    is_trainable=True,
)
class CompactLLMCommandGenerator(LLMBasedCommandGenerator):
    """A single step LLM-based command generator."""

    def __init__(
        self,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        prompt_template: Optional[Text] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            config,
            model_storage,
            resource,
            prompt_template=prompt_template,
            **kwargs,
        )

        # Get the default prompt template based on the model name
        default_command_prompt_template = get_default_prompt_template_based_on_model(
            config, MODEL_PROMPT_MAPPER, DEFAULT_COMMAND_PROMPT_TEMPLATE_FILE_NAME
        )

        # Set the prompt template either from the config or the default prompt template.
        self.prompt_template = prompt_template or get_prompt_template(
            config.get(PROMPT_TEMPLATE_CONFIG_KEY),
            default_command_prompt_template,
        )

        self.trace_prompt_tokens = self.config.get("trace_prompt_tokens", False)
        self.repeat_command_enabled = self.is_repeat_command_enabled()

        # Set the command syntax version to v2
        CommandSyntaxManager.set_syntax_version(CommandSyntaxVersion.v2)

    ### Implementations of LLMBasedCommandGenerator parent
    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        """The component's default config (see parent class for full docstring)."""
        return {
            PROMPT_TEMPLATE_CONFIG_KEY: None,
            USER_INPUT_CONFIG_KEY: None,
            LLM_CONFIG_KEY: None,
            FLOW_RETRIEVAL_KEY: FlowRetrieval.get_default_config(),
        }

    def persist(self) -> None:
        """Persist this component to disk for future loading."""
        self._persist_prompt_template()
        self._persist_config()
        if self.flow_retrieval is not None:
            self.flow_retrieval.persist()

    def _persist_prompt_template(self) -> None:
        """Persist prompt template for future loading."""
        with self._model_storage.write_to(self._resource) as path:
            rasa.shared.utils.io.write_text_file(
                self.prompt_template, path / COMMAND_PROMPT_FILE_NAME
            )

    def _persist_config(self) -> None:
        """Persist config as a source of truth for resolved clients."""
        with self._model_storage.write_to(self._resource) as path:
            rasa.shared.utils.io.dump_obj_as_json_to_file(
                path / LLM_BASED_COMMAND_GENERATOR_CONFIG_FILE, self.config
            )

    @classmethod
    def load(
        cls: Any,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> "CompactLLMCommandGenerator":
        """Loads trained component (see parent class for full docstring)."""
        # Perform health check of the LLM API endpoint
        llm_config = resolve_model_client_config(config.get(LLM_CONFIG_KEY, {}))
        cls.perform_llm_health_check(
            llm_config,
            DEFAULT_LLM_CONFIG,
            "compact_llm_command_generator.load",
            cls.__name__,
        )

        # load prompt template from the model storage.
        prompt_template = cls.load_prompt_template_from_model_storage(
            model_storage, resource, COMMAND_PROMPT_FILE_NAME
        )

        # init base command generator
        command_generator = cls(config, model_storage, resource, prompt_template)
        # load flow retrieval if enabled
        if command_generator.enabled_flow_retrieval:
            command_generator.flow_retrieval = cls.load_flow_retrival(
                command_generator.config, model_storage, resource
            )

        return command_generator

    async def predict_commands(
        self,
        message: Message,
        flows: FlowsList,
        tracker: Optional[DialogueStateTracker] = None,
        **kwargs: Any,
    ) -> List[Command]:
        """Predict commands using the LLM.

        Args:
            message: The message from the user.
            flows: The flows available to the user.
            tracker: The tracker containing the current state of the conversation.
            **kwargs: Keyword arguments for forward compatibility.

        Returns:
            The commands generated by the llm.
        """
        prior_commands = self._get_prior_commands(message)

        if tracker is None or flows.is_empty():
            # cannot do anything if there are no flows or no tracker
            return prior_commands

        if self._should_skip_llm_call(prior_commands, flows, tracker):
            return prior_commands

        try:
            commands = await self._predict_commands(message, flows, tracker)
        except ProviderClientAPIException:
            # if command predictions resulted in API exception
            # "predict" the ErrorCommand
            commands = [ErrorCommand()]
            structlogger.warning(
                "llm_command_generator.predict_commands.api_exception",
                event_info=(
                    "ProviderClientAPIException occurred while predicting commands."
                ),
                commands=commands,
            )

        if not commands and not prior_commands:
            # no commands are parsed or there's an invalid command
            structlogger.warning(
                "llm_command_generator.predict_commands",
                message="No commands were predicted as the LLM response could "
                "not be parsed or the LLM responded with an invalid command."
                "Returning a CannotHandleCommand instead.",
            )
            commands = [CannotHandleCommand()]

        if tracker.has_coexistence_routing_slot:
            # if coexistence feature is used, set the routing slot
            commands += [SetSlotCommand(ROUTE_TO_CALM_SLOT, True)]

        log_llm(
            logger=structlogger,
            log_module=self.__class__.__name__,
            log_event="llm_command_generator.predict_commands.finished",
            commands=commands,
        )

        domain = kwargs.get("domain")
        commands = self._check_commands_against_slot_mappings(commands, tracker, domain)

        return self._check_commands_overlap(prior_commands, commands)

    async def _predict_commands(
        self,
        message: Message,
        flows: FlowsList,
        tracker: Optional[DialogueStateTracker] = None,
    ) -> List[Command]:
        """Predict commands using the LLM.

        Args:
            message: The message from the user.
            flows: The flows available to the user.
            tracker: The tracker containing the current state of the conversation.

        Returns:
            The commands generated by the llm.

        Raises:
            ProviderClientAPIException: If API calls raised an error.
        """
        # retrieve flows
        filtered_flows = await self.filter_flows(message, flows, tracker)

        flow_prompt = self.render_template(message, tracker, filtered_flows, flows)
        log_llm(
            logger=structlogger,
            log_module=self.__class__.__name__,
            log_event="llm_command_generator.predict_commands.prompt_rendered",
            prompt=flow_prompt,
        )

        response = await self.invoke_llm(flow_prompt)
        llm_response = LLMResponse.ensure_llm_response(response)
        # The check for 'None' maintains compatibility with older versions
        # of LLMCommandGenerator. In previous implementations, 'invoke_llm'
        # might return 'None' to indicate a failure to generate actions.
        if llm_response is None or not llm_response.choices:
            structlogger.warning(
                "llm_command_generator.predict_commands.no_actions_generated",
                event_info=(
                    "No actions were generated by the LLM. Returning an ErrorCommand."
                ),
            )
            return [ErrorCommand()]

        action_list = llm_response.choices[0]

        log_llm(
            logger=structlogger,
            log_module=self.__class__.__name__,
            log_event="llm_command_generator.predict_commands.actions_generated",
            action_list=action_list,
        )

        commands = self.parse_commands(action_list, tracker, flows)

        self._update_message_parse_data_for_fine_tuning(message, commands, flow_prompt)
        add_commands_to_message_parse_data(message, self.__class__.__name__, commands)
        add_prompt_to_message_parse_data(
            message=message,
            component_name=self.__class__.__name__,
            prompt_name="command_generator_prompt",
            user_prompt=flow_prompt,
            llm_response=llm_response,
        )

        return commands

    @staticmethod
    def _update_message_parse_data_for_fine_tuning(
        message: Message, commands: List[Command], prompt: str
    ) -> None:
        from rasa.llm_fine_tuning.annotation_module import preparing_fine_tuning_data

        if preparing_fine_tuning_data:
            # Add commands and prompt to the message object in order to create
            # prompt -> commands pairs for fine-tuning
            message.set(
                LLM_COMMANDS,
                [command.as_dict() for command in commands],
                add_to_output=True,
            )
            message.set(LLM_PROMPT, prompt, add_to_output=True)

    @classmethod
    def parse_commands(
        cls, actions: Optional[str], tracker: DialogueStateTracker, flows: FlowsList
    ) -> List[Command]:
        """Parse the actions returned by the llm into intent and entities.

        Args:
            actions: The actions returned by the llm.
            tracker: The tracker containing the current state of the conversation.
            flows: the list of flows

        Returns:
            The parsed commands.
        """
        commands = parse_commands_using_command_parsers(actions, flows)
        if not commands:
            structlogger.debug(
                f"{cls.__class__.__name__}.parse_commands",
                message="No commands were parsed from the LLM actions.",
                actions=actions,
            )

        return commands

    ### Helper methods
    def render_template(
        self,
        message: Message,
        tracker: DialogueStateTracker,
        startable_flows: FlowsList,
        all_flows: FlowsList,
    ) -> str:
        """Render the jinja template to create the prompt for the LLM.

        Args:
            message: The current message from the user.
            tracker: The tracker containing the current state of the conversation.
            startable_flows: The flows startable at this point in time by the user.
            all_flows: all flows present in the assistant

        Returns:
            The rendered prompt template.
        """
        # need to make this distinction here because current step of the
        # top_calling_frame would be the call step, but we need the collect step from
        # the called frame. If no call is active calling and called frame are the same.
        top_calling_frame = top_flow_frame(tracker.stack)
        top_called_frame = top_flow_frame(tracker.stack, ignore_call_frames=False)

        top_flow = top_calling_frame.flow(all_flows) if top_calling_frame else None
        current_step = top_called_frame.step(all_flows) if top_called_frame else None

        flow_slots = self.prepare_current_flow_slots_for_template(
            top_flow, current_step, tracker
        )
        current_slot, current_slot_description = self.prepare_current_slot_for_template(
            current_step
        )
        current_slot_type = None
        current_slot_allowed_values = None
        if current_slot:
            current_slot_type = (
                slot.type_name
                if (slot := tracker.slots.get(current_slot)) is not None
                else None
            )
            current_slot_allowed_values = allowed_values_for_slot(
                tracker.slots.get(current_slot)
            )
        current_conversation = tracker_as_readable_transcript(tracker)
        latest_user_message = sanitize_message_for_prompt(message.get(TEXT))
        current_conversation += f"\nUSER: {latest_user_message}"

        inputs = {
            "available_flows": self.prepare_flows_for_template(
                startable_flows, tracker
            ),
            "current_conversation": current_conversation,
            "flow_slots": flow_slots,
            "current_flow": top_flow.id if top_flow is not None else None,
            "current_slot": current_slot,
            "current_slot_description": current_slot_description,
            "current_slot_type": current_slot_type,
            "current_slot_allowed_values": current_slot_allowed_values,
            "user_message": latest_user_message,
            "is_repeat_command_enabled": self.repeat_command_enabled,
        }

        return self.compile_template(self.prompt_template).render(**inputs)

    def is_repeat_command_enabled(self) -> bool:
        """Check for feature flag"""
        RASA_PRO_BETA_REPEAT_COMMAND_ENV_VAR_NAME = "RASA_PRO_BETA_REPEAT_COMMAND"
        try:
            ensure_beta_feature_is_enabled(
                "Repeat Command",
                env_flag=RASA_PRO_BETA_REPEAT_COMMAND_ENV_VAR_NAME,
            )
        except BetaNotEnabledException:
            return False

        return True

    @classmethod
    def fingerprint_addon(cls: Any, config: Dict[str, Any]) -> Optional[str]:
        """Add a fingerprint for the graph."""
        # Get the default prompt template based on the model name
        default_command_prompt_template = get_default_prompt_template_based_on_model(
            config, MODEL_PROMPT_MAPPER, DEFAULT_COMMAND_PROMPT_TEMPLATE_FILE_NAME
        )
        prompt_template = get_prompt_template(
            config.get(PROMPT_TEMPLATE_CONFIG_KEY),
            default_command_prompt_template,
        )
        llm_config = resolve_model_client_config(
            config.get(LLM_CONFIG_KEY), CompactLLMCommandGenerator.__name__
        )
        embedding_config = resolve_model_client_config(
            config.get(FLOW_RETRIEVAL_KEY, {}).get(EMBEDDINGS_CONFIG_KEY),
            FlowRetrieval.__name__,
        )
        return deep_container_fingerprint(
            [prompt_template, llm_config, embedding_config]
        )



import argparse
from typing import List

import rasa.shared.utils.cli
from rasa.cli import SubParsersAction


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add license parser.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    license_parser = subparsers.add_parser(
        "license",
        parents=parents,
        help="Displays licensing information.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Display licensing information.",
    )
    license_parser.set_defaults(func=display_license_information)


def display_license_information(_: argparse.Namespace) -> None:
    """Display licensing information to stdout."""
    rasa.shared.utils.cli.print_info(
        "By installing and using this software, you agree to be "
        "bound by the terms and conditions of the Developer Terms "
        "available at https://rasa.com/developer-terms. "
        "Please review the Developer Terms carefully before proceeding.\n\n"
        "Rasa Pro relies on several 3rd-party dependencies. "
        "The ones below require a license disclorure:\n",
        PSYCOPG2_LICENSE_DISCLOSURE,
    )


PSYCOPG2_LICENSE_DISCLOSURE = """
psycopg2 & psycopg2-binary
--------------------------

psycopg2 and the LGPL

psycopg2 is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

psycopg2 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

In addition, as a special exception, the copyright holders give permission to link this program with the OpenSSL library (or with modified versions of OpenSSL that use the same license as OpenSSL), and distribute linked combinations including the two.

You must obey the GNU Lesser General Public License in all respects for all of the code used other than OpenSSL. If you modify file(s) with this exception, you may extend this exception to your version of the file(s), but you are not obligated to do so. If you do not wish to do so, delete this exception statement from your version. If you delete this exception statement from all source files in the program, then also delete it here.

You should have received a copy of the GNU Lesser General Public License along with psycopg2 (see the doc/ directory.) If not, see https://www.gnu.org/licenses/. Alternative licenses

The following BSD-like license applies (at your option) to the files following the pattern psycopg/adapter*.{h,c} and psycopg/microprotocol*.{h,c}:

Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:

The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.

Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.

This notice may not be removed or altered from any source distribution.
"""  # noqa: E501



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Set

from rasa.dialogue_understanding.stack.frames import PatternFlowStackFrame
from rasa.shared.constants import RASA_DEFAULT_FLOW_PATTERN_PREFIX
from rasa.shared.core.constants import (
    KEY_ASK_CONFIRM_DIGRESSIONS,
    KEY_BLOCK_DIGRESSIONS,
)

FLOW_PATTERN_HANDLE_DIGRESSIONS = (
    RASA_DEFAULT_FLOW_PATTERN_PREFIX + "handle_digressions"
)


@dataclass
class HandleDigressionsPatternFlowStackFrame(PatternFlowStackFrame):
    """A pattern flow stack frame that gets added if an interruption is completed."""

    flow_id: str = FLOW_PATTERN_HANDLE_DIGRESSIONS
    """The ID of the flow."""
    interrupting_flow_id: str = ""
    """The ID of the flow that interrupted the active flow."""
    interrupted_flow_id: str = ""
    """The name of the active flow that was interrupted."""
    interrupted_step_id: str = ""
    """The ID of the step that was interrupted."""
    ask_confirm_digressions: Set[str] = field(default_factory=set)
    """The set of interrupting flow names to confirm."""
    block_digressions: Set[str] = field(default_factory=set)
    """The set of interrupting flow names to block."""

    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        return FLOW_PATTERN_HANDLE_DIGRESSIONS

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> HandleDigressionsPatternFlowStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        return HandleDigressionsPatternFlowStackFrame(
            frame_id=data["frame_id"],
            step_id=data["step_id"],
            interrupted_step_id=data["interrupted_step_id"],
            interrupted_flow_id=data["interrupted_flow_id"],
            interrupting_flow_id=data["interrupting_flow_id"],
            ask_confirm_digressions=set(data.get(KEY_ASK_CONFIRM_DIGRESSIONS, [])),
            # This attribute must be converted to a set to enable usage
            # of subset `contains` pypred operator in the default pattern
            # conditional branching
            block_digressions=set(data.get(KEY_BLOCK_DIGRESSIONS, [])),
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, HandleDigressionsPatternFlowStackFrame):
            return False
        return (
            self.flow_id == other.flow_id
            and self.interrupted_step_id == other.interrupted_step_id
            and self.interrupted_flow_id == other.interrupted_flow_id
            and self.interrupting_flow_id == other.interrupting_flow_id
            and self.ask_confirm_digressions == other.ask_confirm_digressions
            and self.block_digressions == other.block_digressions
        )

    def as_dict(self) -> Dict[str, Any]:
        """Returns the frame as a dictionary."""
        data = super().as_dict()
        # converting back to list to avoid serialization issues
        data[KEY_ASK_CONFIRM_DIGRESSIONS] = list(self.ask_confirm_digressions)
        data[KEY_BLOCK_DIGRESSIONS] = list(self.block_digressions)
        return data



from __future__ import annotations

from typing import Any, Dict, List, Protocol, runtime_checkable


@runtime_checkable
class RouterClient(Protocol):
    """
    Protocol for a Router client that specifies the interface for interacting
    with the API.
    """

    @classmethod
    def from_config(cls, config: dict) -> RouterClient:
        """
        Initializes the router client with the given configuration.

        This class method should be implemented to parse the given
        configuration and create an instance of an router client.
        """
        ...

    @property
    def config(self) -> Dict:
        """
        Returns the configuration for that the router client is initialized with.

        This property should be implemented to return a dictionary containing
        the client configuration settings for the router client.
        """
        ...

    @property
    def router_settings(self) -> Dict[str, Any]:
        """
        Returns the router settings for the Router client.

        This property should be implemented to return a dictionary containing
        the router settings for the router client.
        """
        ...

    @property
    def model_group_id(self) -> str:
        """
        Returns the model group ID for the Router client.

        This property should be implemented to return the model group ID
        for the router client.
        """
        ...

    @property
    def model_configurations(self) -> List[Dict[str, Any]]:
        """
        Returns the list of model configurations for the Router client.

        This property should be implemented to return the list of model configurations
        for the router client as a list of dictionaries.

        Each dictionary should contain the model configuration.
        Ideally, the `ModelGroupConfig` should parse the model configurations
        and generate this list of dictionaries.
        """
        ...

    @property
    def router_client(self) -> object:
        """
        Returns the instantiated Router client.

        This property should be implemented to return the instantiated
        Router client.
        """
        ...



import structlog
from litellm import validate_environment

from rasa.shared.constants import (
    AWS_ACCESS_KEY_ID_CONFIG_KEY,
    AWS_ACCESS_KEY_ID_ENV_VAR,
    AWS_REGION_NAME_CONFIG_KEY,
    AWS_REGION_NAME_ENV_VAR,
    AWS_SECRET_ACCESS_KEY_CONFIG_KEY,
    AWS_SECRET_ACCESS_KEY_ENV_VAR,
    AWS_SESSION_TOKEN_CONFIG_KEY,
    AWS_SESSION_TOKEN_ENV_VAR,
)
from rasa.shared.exceptions import ProviderClientValidationError
from rasa.shared.providers.embedding._base_litellm_embedding_client import (
    _VALIDATE_ENVIRONMENT_MISSING_KEYS_KEY,
)

structlogger = structlog.get_logger()


def validate_aws_setup_for_litellm_clients(
    litellm_model_name: str, litellm_call_kwargs: dict, source_log: str
) -> None:
    """Validates the AWS setup for LiteLLM clients to ensure all required
    environment variables or corresponding call kwargs are set.

    Args:
        litellm_model_name (str): The name of the LiteLLM model being validated.
        litellm_call_kwargs (dict): Additional keyword arguments passed to the client,
            which may include configuration values for AWS credentials.
        source_log (str): The source log identifier for structured logging.

    Raises:
        ProviderClientValidationError: If any required AWS environment variable
            or corresponding configuration key is missing.
    """

    # Mapping of environment variable names to their corresponding config keys
    envs_to_args = {
        AWS_ACCESS_KEY_ID_ENV_VAR: AWS_ACCESS_KEY_ID_CONFIG_KEY,
        AWS_SECRET_ACCESS_KEY_ENV_VAR: AWS_SECRET_ACCESS_KEY_CONFIG_KEY,
        AWS_REGION_NAME_ENV_VAR: AWS_REGION_NAME_CONFIG_KEY,
        AWS_SESSION_TOKEN_ENV_VAR: AWS_SESSION_TOKEN_CONFIG_KEY,
    }

    # Validate the environment setup for the model
    validation_info = validate_environment(litellm_model_name)
    missing_environment_variables = validation_info.get(
        _VALIDATE_ENVIRONMENT_MISSING_KEYS_KEY, []
    )
    # Filter out missing environment variables that have been set trough arguments
    # in extra parameters
    missing_environment_variables = [
        missing_env_var
        for missing_env_var in missing_environment_variables
        if litellm_call_kwargs.get(envs_to_args.get(missing_env_var)) is None
    ]

    if missing_environment_variables:
        missing_environment_details = [
            (
                f"'{missing_env_var}' environment variable or "
                f"'{envs_to_args.get(missing_env_var)}' config key"
            )
            for missing_env_var in missing_environment_variables
        ]
        event_info = (
            f"The following environment variables or configuration keys are "
            f"missing: "
            f"{', '.join(missing_environment_details)}. "
            f"These settings are required for API calls."
        )
        structlogger.error(
            f"{source_log}.validate_aws_environment_variables",
            event_info=event_info,
            missing_environment_variables=missing_environment_variables,
        )
        raise ProviderClientValidationError(event_info)
